# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
import importlib
import itertools
import re
import typing

from mlrun.errors import MLRunInvalidArgumentError
from mlrun.utils import logger


class TypeHintUtils:
    """
    Static class for utilities functions to process type hints.
    """

    @staticmethod
    def is_typing_type(type_hint: type) -> bool:
        """
        Check whether a given type is a type hint from one of the modules `typing` and `types`. The function will return
        True for generic type aliases also, meaning Python 3.9's new hinting feature that includes hinting like
        `list[int]` instead of `typing.List[int]`.

        :param type_hint: The type to check.

        :return: True if the type hint from `typing` / `types` and False otherwise.
        """
        # A type hint should be one of the based typing classes, meaning it will have "typing" as its module. Some
        # typing classes are considered a type (like `TypeVar`) so we check their type as well. The only case "types"
        # will be a module is for generic aliases like `list[int]`.
        return (type_hint.__module__ == "typing") or (
            type(type_hint).__module__ in ["typing", "types"]
        )

    @staticmethod
    def parse_type_hint(type_hint: typing.Union[type, str]) -> type:
        """
        Parse a given type hint from string to its actual hinted type class object. The string must be one of the
        following:

        * Python builtin type - for example: `tuple`, `list`, `set`, `dict` and `bytearray`.
        * Full module import path. An alias (if `import pandas as pd is used`, the type hint cannot be `pd.DataFrame`)
          is not allowed.

        The type class on its own (like `DataFrame`) cannot be used as the scope of this function is not the same as the
        handler itself, hence modules and objects that were imported in the handler's scope are not available. This is
        the same reason import aliases cannot be used as well.

        If the provided type hint is not a string, it will simply be returned as is.

        :param type_hint: The type hint to parse.

        :return: The hinted type.

        :raise MLRunInvalidArgumentError: In case the type hint is not following the 2 options mentioned above.
        """
        if not isinstance(type_hint, str):
            return type_hint

        # Validate the type hint is a valid module path:
        if not bool(
            re.fullmatch(
                r"([a-zA-Z_][a-zA-Z0-9_]*\.)*[a-zA-Z_][a-zA-Z0-9_]*", type_hint
            )
        ):
            raise MLRunInvalidArgumentError(
                f"Invalid type hint. An input type hint must be a valid python class name or its module import path. "
                f"For example: 'list', 'pandas.DataFrame', 'numpy.ndarray', 'sklearn.linear_model.LinearRegression'. "
                f"Type hint given: '{type_hint}'."
            )

        # Look for a builtin type (rest of the builtin types like `int`, `str`, `float` should be treated as results,
        # hence not given as an input to an MLRun function, but as a parameter):
        builtin_types = {
            builtin_name: builtin_type
            for builtin_name, builtin_type in builtins.__dict__.items()
            if isinstance(builtin_type, type)
        }
        if type_hint in builtin_types:
            return builtin_types[type_hint]

        # If it's not a builtin, its should have a full module path, meaning at least one '.' to separate the module and
        # the class. If it doesn't, we will try to get the class from the main module:
        if "." not in type_hint:
            logger.warn(
                f"The type hint string given '{type_hint}' is not a `builtins` python type. MLRun will try to look for "
                f"it in the `__main__` module instead."
            )
            try:
                return TypeHintUtils.parse_type_hint(type_hint=f"__main__.{type_hint}")
            except MLRunInvalidArgumentError:
                raise MLRunInvalidArgumentError(
                    f"MLRun tried to get the type hint '{type_hint}' but it can't as it is not a valid builtin Python "
                    f"type (one of `list`, `dict`, `str`, `int`, etc.) nor a locally declared type (from the "
                    f"`__main__` module). Pay attention using only the type as string is not allowed as the handler's "
                    f"scope is different than MLRun's. To properly give a type hint as string, please specify the full "
                    f"module path without aliases. For example: do not use `DataFrame` or `pd.DataFrame`, use "
                    f"`pandas.DataFrame`."
                )

        # Import the module to receive the hinted type:
        try:
            # Get the module path and the type class (If we'll wish to support inner classes, the `rsplit` won't work):
            module_path, type_hint = type_hint.rsplit(".", 1)
            # Replace alias if needed (alias assumed to be imported already, hence we look in globals):
            # For example:
            # If in handler scope there was `import A.B.C as abc` and user gave a type hint "abc.Something" then:
            # `module_path[0]` will be equal to "abc". Then, because it is an alias, it will appear in the globals, so
            # we'll replace the alias with the full module name in order to import the module.
            module_path = module_path.split(".")
            if module_path[0] in globals():
                module_path[0] = globals()[module_path[0]].__name__
            module_path = ".".join(module_path)
            # Import the module:
            module = importlib.import_module(module_path)
            # Get the class type from the module:
            type_hint = getattr(module, type_hint)
        except ModuleNotFoundError as module_not_found_error:
            # May be raised from `importlib.import_module` in case the module does not exist.
            raise MLRunInvalidArgumentError(
                f"MLRun tried to get the type hint '{type_hint}' but the module '{module_path}' cannot be imported. "
                f"Keep in mind that using alias in the module path (meaning: import module as alias) is not allowed. "
                f"If the module path is correct, please make sure the module package is installed in the python "
                f"interpreter."
            ) from module_not_found_error
        except AttributeError as attribute_error:
            # May be raised from `getattr(module, type_hint)` in case the class type cannot be imported directly from
            # the imported module.
            raise MLRunInvalidArgumentError(
                f"MLRun tried to get the type hint '{type_hint}' from the module '{module.__name__}' but it seems it "
                f"doesn't exist. Make sure the class can be imported from the module with the exact module path you "
                f"passed. Notice inner classes (a class inside of a class) are not supported."
            ) from attribute_error

        return type_hint

    @staticmethod
    def is_matching(
        object_type: type,
        type_hint: typing.Union[type, set[type]],
        include_subclasses: bool = True,
        reduce_type_hint: bool = True,
    ) -> bool:
        """
        Check if the given object type match the given hint.

        :param object_type:        The object type to match with the type hint.
        :param type_hint:          The hint to match with. Can be given as a set resulted from a reduced hint.
        :param include_subclasses: Whether to mark a subclass as valid match. Default to True.
        :param reduce_type_hint:   Whether to reduce the type hint to match with its reduced hints.

        :return: True if the object type match the type hint and False otherwise.
        """
        # Wrap in a set if provided a single type hint:
        type_hint = {type_hint} if not isinstance(type_hint, set) else type_hint

        # Try to match the object type to one of the hints:
        while len(type_hint) > 0:
            for hint in type_hint:
                # Subclass check can be made only on actual object types (not typing module types):
                if (
                    not TypeHintUtils.is_typing_type(type_hint=object_type)
                    and not TypeHintUtils.is_typing_type(type_hint=hint)
                    and include_subclasses
                    and issubclass(object_type, hint)
                ):
                    return True
                if object_type == hint:
                    return True
            # See if needed to reduce, if not end on first iteration:
            if not reduce_type_hint:
                break
            type_hint = TypeHintUtils.reduce_type_hint(type_hint=type_hint)
        return False

    @staticmethod
    def reduce_type_hint(
        type_hint: typing.Union[type, set[type]],
    ) -> set[type]:
        """
        Reduce a type hint (or a set of type hints) using the `_reduce_type_hint` function.

        :param type_hint: The type hint to reduce.

        :return: The reduced type hints set or an empty set if the type hint could not be reduced.
        """
        # Wrap in a set if provided a single type hint:
        type_hints = {type_hint} if not isinstance(type_hint, set) else type_hint

        # Iterate over the type hints and reduce each one:
        return set(
            itertools.chain(
                *[
                    TypeHintUtils._reduce_type_hint(type_hint=type_hint)
                    for type_hint in type_hints
                ]
            )
        )

    @staticmethod
    def _reduce_type_hint(type_hint: type) -> list[type]:
        """
        Reduce a type hint. If the type hint is a `typing` module, it will be reduced to its original hinted types. For
        example: `typing.Union[int, float, typing.List[int]]` will return `[int, float, List[int]]` and
        `typing.List[int]` will return `[list]`. Regular type hints - Python object types cannot be reduced as they are
        already a core type.

        If a type hint cannot be reduced, an empty list will be returned.

        :param type_hint: The type hint to reduce.

        :return: The reduced type hint as list of hinted types or an empty list if the type hint could not be reduced.
        """
        # If it's not a typing type (meaning it's an actual object type) then we can't reduce it further:
        if not TypeHintUtils.is_typing_type(type_hint=type_hint):
            return []

        # If it's a type var, take its constraints (e.g. A = TypeVar("A", int, str) meaning an object of type A should
        # be an integer or a string). If it doesn't have constraints, return an empty list:
        if isinstance(type_hint, typing.TypeVar):
            if len(type_hint.__constraints__) == 0:
                return []
            return list(type_hint.__constraints__)

        # If it's a forward reference, we will try to import the reference:
        if isinstance(type_hint, typing.ForwardRef):
            try:
                # ForwardRef is initialized with the string type it represents and optionally a module path, so we
                # construct a full module path and try to parse it:
                arg = type_hint.__forward_arg__
                if type_hint.__forward_module__:
                    arg = f"{type_hint.__forward_module__}.{arg}"
                return [TypeHintUtils.parse_type_hint(type_hint=arg)]
            except (
                MLRunInvalidArgumentError
            ):  # May be raised from `TypeHintUtils.parse_type_hint`
                logger.warn(
                    f"Could not reduce the type hint '{type_hint}' as it is a forward reference to a class without "
                    f"it's full module path. To enable importing forward references, please provide the full module "
                    f"path to them. For example: use `ForwardRef('pandas.DataFrame')` instead of "
                    f"`ForwardRef('DataFrame')`."
                )
                return []

        # Get the origin of the typing type. An origin is the subscripted typing type (origin of Union[str, int] is
        # Union). The origin can be one of Callable, Tuple, Union, Literal, Final, ClassVar, Annotated or the actual
        # type alias (e.g. origin of List[int] is list):
        origin = typing.get_origin(type_hint)

        # If the typing type has no origin (e.g. None is returned), we cannot reduce it, so we return an empty list:
        if origin is None:
            return []

        # If the origin is a type of one of `builtins`, `contextlib` or `collections` (for example: List's origin is
        # list) then we can be sure there is nothing to reduce as it's a regular type:
        if not TypeHintUtils.is_typing_type(type_hint=origin):
            return [origin]

        # Get the type's subscriptions - arguments, in order to reduce it to them (we know for sure there are arguments,
        # otherwise origin would have been None):
        args = typing.get_args(type_hint)

        # Return the reduced type as its arguments according to the origin:
        if origin is typing.Callable:
            # A callable cannot be reduced to its arguments, so we'll return the origin - Callable:
            return [typing.Callable]
        if origin is typing.Literal:
            # Literal arguments are not types, but values. So we'll take the types of the values as the reduced type:
            return [type(arg) for arg in args]
        if origin is typing.Union:
            # A union is reduced to its arguments:
            return list(args)
        if origin is typing.Annotated:
            # Annotated is used to describe (add metadata to) a type, so we take the first argument (the type the
            # metadata is being added to):
            return [args[0]]
        if origin is typing.Final or origin is typing.ClassVar:
            # Both Final and ClassVar takes only one argument - the type:
            return [args[0]]

        # For Generic types we return an empty list:
        return []



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# this module is WIP
from enum import Enum

import pyarrow
from pyarrow.lib import TimestampType


class ValueType(str, Enum):
    """Feature value type. Used to define data types in Feature Tables."""

    UNKNOWN = ""
    BOOL = "bool"
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int"
    INT128 = "int128"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    UINT128 = "uint128"
    FLOAT16 = "float16"
    FLOAT = "float32"
    DOUBLE = "float"
    BFLOAT16 = "bfloat16"
    BYTES = "bytes"
    STRING = "str"
    DATETIME = "datetime"
    LIST = "List"
    BYTES_LIST = "List[bytes]"
    STRING_LIST = "List[string]"
    INT32_LIST = "List[int32]"
    INT64_LIST = "List[int]"
    DOUBLE_LIST = "List[float]"
    FLOAT_LIST = "List[float32]"
    BOOL_LIST = "List[bool]"
    Tuple = "Tuple"


def pd_schema_to_value_type(value):
    type_map = {
        "integer": ValueType.INT64,
        "string": ValueType.STRING,
        "number": ValueType.DOUBLE,
        "datetime": ValueType.DATETIME,
        "boolean": ValueType.BOOL,
        "any": ValueType.STRING,
        "duration": ValueType.INT64,
    }
    return type_map[value]


def pa_type_to_value_type(type_):
    # To catch timestamps with timezones. This also catches timestamps with different units
    if isinstance(type_, TimestampType):
        return ValueType.DATETIME

    # pandas category type translates to pyarrow DictionaryType
    # we need to unpack the value type (ML-7868)
    if isinstance(type_, pyarrow.DictionaryType):
        type_ = type_.value_type

    type_map = {
        pyarrow.bool_(): ValueType.BOOL,
        pyarrow.int64(): ValueType.INT64,
        pyarrow.int32(): ValueType.INT32,
        pyarrow.float32(): ValueType.FLOAT,
        pyarrow.float64(): ValueType.DOUBLE,
    }
    return type_map.get(type_, ValueType.STRING)


def python_type_to_value_type(value_type):
    type_name = (
        value_type.__name__ if hasattr(value_type, "__name__") else str(value_type)
    )
    type_map = {
        "int": ValueType.INT64,
        "str": ValueType.STRING,
        "float": ValueType.DOUBLE,
        "bytes": ValueType.BYTES,
        "float64": ValueType.DOUBLE,
        "Float64": ValueType.DOUBLE,
        "float32": ValueType.FLOAT,
        "int64": ValueType.INT64,
        "Int64": ValueType.INT64,
        "uint64": ValueType.INT64,
        "int32": ValueType.INT32,
        "uint32": ValueType.INT32,
        "uint8": ValueType.INT32,
        "int8": ValueType.INT32,
        "bool": ValueType.BOOL,
        "timedelta": ValueType.INT64,
        "datetime64[ns]": ValueType.INT64,
        "datetime64[ns, tz]": ValueType.INT64,
        "category": ValueType.STRING,
        "list": ValueType.LIST,
        "tuple": ValueType.Tuple,
    }

    if type_name in type_map:
        return type_map[type_name]


def spark_to_value_type(data_type):
    type_map = {
        "int": ValueType.INT64,
        "bigint": ValueType.INT64,
        "double": ValueType.DOUBLE,
        "boolean": ValueType.BOOL,
        "timestamp": ValueType.DATETIME,
        "timestamp_ntz": ValueType.DATETIME,
        "string": ValueType.STRING,
        "array": "list",
        "map": "dict",
    }
    if data_type in type_map:
        return type_map[data_type]
    return data_type


def gbq_to_pandas_dtype(gbq_type):
    import pandas as pd

    if isinstance(gbq_type, TimestampType):
        return ValueType.DATETIME

    type_map = {
        "STRING": "object",
        "BOOL": "bool",
        "FLOAT": "float64",
        "INTEGER": pd.Int64Dtype(),
        "TIMESTAMP": "datetime64[ns, UTC]",
    }
    return type_map.get(gbq_type, "object")


class InferOptions:
    Null = 0
    Entities = 1
    Features = 2
    Index = 4
    Stats = 8
    Histogram = 16
    Preview = 32

    @staticmethod
    def schema():
        return InferOptions.Entities + InferOptions.Features + InferOptions.Index

    @staticmethod
    def all_stats():
        return InferOptions.Stats + InferOptions.Histogram + InferOptions.Preview

    @staticmethod
    def all():
        return (
            InferOptions.schema()
            + InferOptions.Stats
            + InferOptions.Histogram
            + InferOptions.Preview
        )

    @staticmethod
    def default():
        return InferOptions.all()

    @staticmethod
    def get_common_options(one, two):
        return one & two



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
import json
import typing
import warnings
from datetime import datetime
from time import sleep

import inflection
import nuclio
import nuclio.utils
import requests
import semver
from aiohttp.client import ClientSession
from kubernetes import client
from nuclio.deploy import find_dashboard_url, get_deploy_status
from nuclio.triggers import V3IOStreamTrigger

import mlrun.common.constants
import mlrun.db
import mlrun.errors
import mlrun.k8s_utils
import mlrun.utils
import mlrun.utils.helpers
from mlrun.common.schemas import AuthInfo
from mlrun.config import config as mlconf
from mlrun.errors import err_to_str
from mlrun.lists import RunList
from mlrun.model import RunObject
from mlrun.platforms.iguazio import (
    parse_path,
    split_path,
)
from mlrun.runtimes.base import FunctionStatus, RunError
from mlrun.runtimes.mounts import VolumeMount, mount_v3io, v3io_cred
from mlrun.runtimes.pod import KubeResource, KubeResourceSpec
from mlrun.runtimes.utils import get_item_name, log_std
from mlrun.utils import get_in, logger, update_in
from mlrun_pipelines.common.ops import deploy_op

SENSITIVE_PATHS_IN_TRIGGER_CONFIG = {
    "password",
    "secret",
    "attributes/password",
    "attributes/accesskeyid",
    "attributes/secretaccesskey",
    "attributes/cacert",
    "attributes/accesskey",
    "attributes/accesscertificate",
    "attributes/sasl/password",
    "attributes/sasl/oauth/clientsecret",
}


def validate_nuclio_version_compatibility(*min_versions):
    """
    :param min_versions: Valid minimum version(s) required, assuming no 2 versions has equal major and minor.
    """
    return mlrun.utils.helpers.validate_component_version_compatibility(
        "nuclio", *min_versions
    )


def min_nuclio_versions(*versions):
    def decorator(function):
        def wrapper(*args, **kwargs):
            if validate_nuclio_version_compatibility(*versions):
                return function(*args, **kwargs)

            if function.__name__ == "__init__":
                name = inflection.titleize(function.__qualname__.split(".")[0])
            else:
                name = function.__qualname__

            message = (
                f"'{name}' function requires Nuclio v{' or v'.join(versions)} or higher"
            )
            raise mlrun.errors.MLRunIncompatibleVersionError(message)

        return wrapper

    return decorator


class NuclioSpec(KubeResourceSpec):
    _dict_fields = KubeResourceSpec._dict_fields + [
        "min_replicas",
        "max_replicas",
        "config",
        "base_spec",
        "no_cache",
        "source",
        "function_kind",
        "readiness_timeout",
        "readiness_timeout_before_failure",
        "function_handler",
        "nuclio_runtime",
        "base_image_pull",
        "service_type",
        "add_templated_ingress_host_mode",
        "disable_default_http_trigger",
    ]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        entry_points=None,
        description=None,
        replicas=None,
        min_replicas=None,
        max_replicas=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        config=None,
        base_spec=None,
        no_cache=None,
        source=None,
        image_pull_policy=None,
        function_kind=None,
        build=None,
        service_account=None,
        readiness_timeout=None,
        readiness_timeout_before_failure=None,
        default_handler=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        disable_auto_mount=False,
        priority_class_name=None,
        pythonpath=None,
        workdir=None,
        image_pull_secret=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        service_type=None,
        add_templated_ingress_host_mode=None,
        state_thresholds=None,
        disable_default_http_trigger=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            build=build,
            entry_points=entry_points,
            description=description,
            default_handler=default_handler,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            disable_auto_mount=disable_auto_mount,
            priority_class_name=priority_class_name,
            pythonpath=pythonpath,
            workdir=workdir,
            image_pull_secret=image_pull_secret,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
        )

        self.base_spec = base_spec or {}
        self.function_kind = function_kind
        self.source = source or ""
        self.config = config or {}
        self.function_handler = None
        self.nuclio_runtime = None
        self.no_cache = no_cache
        self.readiness_timeout = readiness_timeout
        self.readiness_timeout_before_failure = readiness_timeout_before_failure
        self.service_type = service_type
        self.add_templated_ingress_host_mode = add_templated_ingress_host_mode
        self.state_thresholds = None  # not supported in nuclio

        self.min_replicas = min_replicas or 1
        self.max_replicas = max_replicas or 4

        self.disable_default_http_trigger = disable_default_http_trigger

        # When True it will set Nuclio spec.noBaseImagesPull to False (negative logic)
        # indicate that the base image should be pulled from the container registry (not cached)
        self.base_image_pull = False

    def generate_nuclio_volumes(self):
        nuclio_volumes = []
        volume_with_volume_mounts_names = set()
        for volume_mount in self._volume_mounts.values():
            volume_name = get_item_name(volume_mount, "name")
            if volume_name not in self._volumes:
                raise ValueError(
                    f"Found volume mount without a volume. name={volume_name}"
                )
            volume_with_volume_mounts_names.add(volume_name)
            nuclio_volumes.append(
                {"volume": self._volumes[volume_name], "volumeMount": volume_mount}
            )

        volumes_without_volume_mounts = (
            volume_with_volume_mounts_names.symmetric_difference(self._volumes.keys())
        )
        if volumes_without_volume_mounts:
            raise ValueError(
                f"Found volumes without volume mounts. names={volumes_without_volume_mounts}"
            )

        return nuclio_volumes


class NuclioStatus(FunctionStatus):
    def __init__(
        self,
        state=None,
        nuclio_name=None,
        address=None,
        internal_invocation_urls=None,
        external_invocation_urls=None,
        build_pod=None,
        container_image=None,
    ):
        super().__init__(state, build_pod)

        self.nuclio_name = nuclio_name

        # exists on nuclio >= 1.6.x
        # infers the function invocation urls
        self.internal_invocation_urls = internal_invocation_urls or []
        self.external_invocation_urls = external_invocation_urls or []

        # still exists for backwards compatability reasons.
        # on latest Nuclio (>= 1.6.x) versions, use external_invocation_urls / internal_invocation_urls instead
        self.address = address

        # the name of the image that was built and pushed to the registry, and used by the nuclio function
        self.container_image = container_image


class RemoteRuntime(KubeResource):
    kind = "remote"
    _is_nested = False
    _mock_server = None

    @property
    def spec(self) -> NuclioSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", NuclioSpec)

    @property
    def status(self) -> NuclioStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", NuclioStatus)

    def pre_deploy_validation(self):
        if self.metadata.tag:
            mlrun.utils.validate_tag_name(self.metadata.tag, "function.metadata.tag")

    def mask_sensitive_data_in_config(self):
        if not self.spec.config:
            return {}

        raw_config = copy.deepcopy(self.spec.config)

        for key, value in self.spec.config.items():
            if key.startswith("spec.triggers"):
                trigger_name = key.split(".")[-1]

                for path in SENSITIVE_PATHS_IN_TRIGGER_CONFIG:
                    # Handle nested keys
                    nested_keys = path.split("/")
                    target = value
                    for sub_key in nested_keys[:-1]:
                        target = target.get(sub_key, {})

                    last_key = nested_keys[-1]
                    if last_key in target:
                        sensitive_field = target[last_key]
                        if sensitive_field.startswith(
                            mlrun.model.Credentials.secret_reference_prefix
                        ):
                            # already masked
                            continue
                        target[last_key] = (
                            f"{mlrun.model.Credentials.secret_reference_prefix}/spec/triggers/{trigger_name}/{path}"
                        )

        return raw_config

    def set_config(self, key, value):
        self.spec.config[key] = value
        return self

    def with_annotations(self, annotations: dict):
        """set a key/value annotations for function"""

        self.spec.base_spec.setdefault("metadata", {})
        self.spec.base_spec["metadata"].setdefault("annotations", {})
        for key, value in annotations.items():
            self.spec.base_spec["metadata"]["annotations"][key] = str(value)

        return self

    def add_trigger(self, name, spec):
        """add a nuclio trigger object/dict

        :param name: trigger name
        :param spec: trigger object or dict
        """
        if hasattr(spec, "to_dict"):
            spec = spec.to_dict()

        self._validate_triggers(spec)

        spec["name"] = name
        self.spec.config[f"spec.triggers.{name}"] = spec
        return self

    def _validate_triggers(self, spec):
        # ML-7763 / NUC-233
        min_nuclio_version = "1.13.12"
        if (
            mlconf.nuclio_version
            and mlconf.nuclio_version != "unstable"
            and semver.VersionInfo.parse(mlconf.nuclio_version)
            < semver.VersionInfo.parse(min_nuclio_version)
        ):
            explicit_ack_enabled = False
            num_triggers = 0
            trigger_name = spec.get("name", "UNKNOWN")
            for key, config in [(f"spec.triggers.{trigger_name}", spec)] + list(
                self.spec.config.items()
            ):
                if key.startswith("spec.triggers."):
                    num_triggers += 1
                    explicit_ack_enabled = (
                        config.get("explicitAckMode", "disable") != "disable"
                    )

            if num_triggers > 1 and explicit_ack_enabled:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Multiple triggers cannot be used in conjunction with explicit ack. "
                    f"Please upgrade to nuclio {min_nuclio_version} or newer."
                )

    def with_source_archive(
        self,
        source,
        workdir=None,
        handler=None,
        runtime="",
    ):
        """Load nuclio function from remote source

        Note: remote source may require credentials, those can be stored in the project secrets or passed
        in the function.deploy() using the builder_env dict, see the required credentials per source:

        - v3io - "V3IO_ACCESS_KEY".
        - git - "GIT_USERNAME", "GIT_PASSWORD".
        - AWS S3 - "AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY" or "AWS_SESSION_TOKEN".

        :param source: a full path to the nuclio function source (code entry) to load the function from
        :param handler: a path to the function's handler, including path inside archive/git repo
        :param workdir: working dir  relative to the archive root (e.g. 'subdir')
        :param runtime: (optional) the runtime of the function (defaults to mlrun.mlconf.default_nuclio_runtime)

        :Examples:

            git::

                fn.with_source_archive(
                    "git://github.com/org/repo#my-branch",
                    handler="main:handler",
                    workdir="path/inside/repo",
                )

            s3::

                fn.spec.nuclio_runtime = "golang"
                fn.with_source_archive(
                    "s3://my-bucket/path/in/bucket/my-functions-archive",
                    handler="my_func:Handler",
                    workdir="path/inside/functions/archive",
                    runtime="golang",
                )
        """
        self.spec.build.source = source

        code = (
            self.spec.build.functionSourceCode if hasattr(self.spec, "build") else None
        )
        if code:
            # Warn and clear any inline code so the archive is actually used
            logger.warning(
                "Cannot specify both code and source archive. Removing the code so the provided "
                "source archive will be used instead."
            )
            self.spec.build.functionSourceCode = None

        # update handler in function_handler if needed
        if handler:
            self.spec.function_handler = handler
        if workdir:
            self.spec.workdir = workdir
        if runtime:
            self.spec.nuclio_runtime = runtime

        return self

    def with_v3io(self, local="", remote=""):
        """Add v3io volume to the function

        :param local: local path (mount path inside the function container)
        :param remote: v3io path
        """
        if local and remote:
            self.apply(
                mount_v3io(
                    remote=remote, volume_mounts=[VolumeMount(path=local, sub_path="")]
                )
            )
        else:
            self.apply(v3io_cred())
        return self

    def with_http(
        self,
        workers: typing.Optional[int] = 8,
        port: typing.Optional[int] = None,
        host: typing.Optional[str] = None,
        paths: typing.Optional[list[str]] = None,
        canary: typing.Optional[float] = None,
        secret: typing.Optional[str] = None,
        worker_timeout: typing.Optional[int] = None,
        gateway_timeout: typing.Optional[int] = None,
        trigger_name: typing.Optional[str] = None,
        annotations: typing.Optional[typing.Mapping[str, str]] = None,
        extra_attributes: typing.Optional[typing.Mapping[str, str]] = None,
    ):
        """update/add nuclio HTTP trigger settings

        Note: gateway timeout is the maximum request time before an error is returned, while the worker timeout
        if the max time a request will wait for until it will start processing, gateway_timeout must be greater than
        the worker_timeout.

        :param workers:    number of worker processes (default=8). set 0 to use Nuclio's default workers count
        :param port:       TCP port to listen on. by default, nuclio will choose a random port as long as
                           the function service is NodePort. if the function service is ClusterIP, the port
                           is ignored.
        :param host:       Ingress hostname
        :param paths:      list of Ingress sub paths
        :param canary:     k8s ingress canary (% traffic value between 0 and 100)
        :param secret:     k8s secret name for SSL certificate
        :param worker_timeout:  worker wait timeout in sec (how long a message should wait in the worker queue
                                before an error is returned)
        :param gateway_timeout: nginx ingress timeout in sec (request timeout, when will the gateway return an error)
        :param trigger_name:    alternative nuclio trigger name
        :param annotations:     key/value dict of ingress annotations
        :param extra_attributes: key/value dict of extra nuclio trigger attributes
        :return: function object (self)
        """
        if self.disable_default_http_trigger:
            logger.warning(
                "Adding HTTP trigger despite the default HTTP trigger creation being disabled"
            )

        annotations = annotations or {}
        if worker_timeout:
            gateway_timeout = gateway_timeout or (worker_timeout + 60)
        if workers is None:
            workers = 0
        if gateway_timeout:
            if worker_timeout and worker_timeout >= gateway_timeout:
                raise ValueError(
                    "gateway timeout must be greater than the worker timeout"
                )
            mlrun.runtimes.utils.enrich_gateway_timeout_annotations(
                annotations, gateway_timeout
            )

        trigger = nuclio.HttpTrigger(
            workers=workers,
            port=port,
            host=host,
            paths=paths,
            canary=canary,
            secret=secret,
            annotations=annotations,
            extra_attributes=extra_attributes,
        )
        if worker_timeout:
            trigger._struct["workerAvailabilityTimeoutMilliseconds"] = (
                worker_timeout
            ) * 1000
        self.add_trigger(trigger_name or "http", trigger)
        return self

    def from_image(self, image):
        """
        Deploy the function with an existing nuclio processor image.

        :param image: image name
        """
        config = nuclio.config.new_config()
        update_in(
            config,
            "spec.handler",
            self.spec.function_handler or "main:handler",
        )
        update_in(config, "spec.image", image)
        update_in(config, "spec.build.codeEntryType", "image")
        self.spec.base_spec = config

    def add_v3io_stream_trigger(
        self,
        stream_path,
        name="stream",
        group="serving",
        seek_to="earliest",
        shards=1,
        extra_attributes=None,
        ack_window_size=None,
        **kwargs,
    ):
        """add v3io stream trigger to the function

        :param stream_path:    v3io stream path (e.g. 'v3io:///projects/myproj/stream1')
        :param name:           trigger name
        :param group:          consumer group
        :param seek_to:        start seek from: "earliest", "latest", "time", "sequence"
        :param shards:         number of shards (used to set number of replicas)
        :param extra_attributes: key/value dict with extra trigger attributes
        :param ack_window_size:  stream ack window size (the consumer group will be updated with the
                                 event id - ack_window_size, on failure the events in the window will be retransmitted)
        :param kwargs:         extra V3IOStreamTrigger class attributes
        """
        endpoint = None
        if "://" in stream_path:
            endpoint, stream_path = parse_path(stream_path, suffix="")

        # verify v3io stream trigger name is valid
        mlrun.utils.helpers.validate_v3io_stream_consumer_group(group)

        if "consumer_group" in kwargs:
            logger.warning(
                "'consumer_group' in kwargs will be ignored. Use group parameter instead."
            )

        container, path = split_path(stream_path)
        shards = shards or 1
        extra_attributes = extra_attributes or {}
        if ack_window_size:
            extra_attributes["ackWindowSize"] = ack_window_size

        access_key = kwargs.pop("access_key", None)
        if not access_key:
            access_key = self._resolve_v3io_access_key()
        engine = "sync"
        explicit_ack_mode = kwargs.pop("explicit_ack_mode", None)
        if self.spec and hasattr(self.spec, "graph"):
            engine = getattr(self.spec.graph, "engine", None) or engine

        if mlrun.mlconf.is_explicit_ack_enabled() and engine == "async":
            explicit_ack_mode = explicit_ack_mode or "explicitOnly"

        self.add_trigger(
            name,
            V3IOStreamTrigger(
                name=name,
                container=container,
                path=path[1:],
                consumer_group=group,
                seek_to=seek_to,
                webapi=endpoint or "http://v3io-webapi:8081",
                extra_attributes=extra_attributes,
                read_batch_size=256,
                access_key=access_key,
                explicit_ack_mode=explicit_ack_mode,
                **kwargs,
            ),
        )
        if self.spec.min_replicas != shards or self.spec.max_replicas != shards:
            logger.warning(f"Setting function replicas to {shards}")
            self.spec.min_replicas = shards
            self.spec.max_replicas = shards

    def deploy(
        self,
        project="",
        tag="",
        verbose=False,
        builder_env: typing.Optional[dict] = None,
        force_build: bool = False,
    ):
        """Deploy the nuclio function to the cluster

        :param project:    project name
        :param tag:        function tag
        :param verbose:    set True for verbose logging
        :param builder_env: env vars dict for source archive config/credentials e.g. builder_env={"GIT_TOKEN": token}
        :param force_build: set True for force building the image
        """

        old_http_session = getattr(self, "_http_session", None)
        if old_http_session:
            # ensure existing http session is terminated prior to (re)deploy to ensure that a connection to an old
            # replica will not be reused
            old_http_session.close()
            self._http_session = None

        verbose = verbose or self.verbose
        if verbose:
            self.set_env("MLRUN_LOG_LEVEL", "DEBUG")
        if project:
            self.metadata.project = project
        if tag:
            self.metadata.tag = tag

        mlrun.utils.helpers.validate_function_name(self.metadata.name)

        # Attempt auto-mounting, before sending to remote build
        self.try_auto_mount_based_on_config()
        self._fill_credentials()
        db = self._get_db()
        logger.info("Starting remote function deploy")
        data = db.deploy_nuclio_function(func=self, builder_env=builder_env)
        self.status = data["data"].get("status")

        # Extract the spec to avoid overwriting server-side updates during the later save in
        # _enrich_command_from_status.
        self.spec = data["data"].get("spec")

        self._update_credentials_from_remote_build(data["data"])

        # when a function is deployed, we wait for it to be ready by default
        # this also means that the function object will be updated with the function status
        self._wait_for_function_deployment(db, verbose=verbose)
        # check if there are any background tasks related to creating model endpoints
        model_endpoints_creation_background_tasks = (
            mlrun.common.schemas.BackgroundTaskList(
                **data.pop("background_tasks", {"background_tasks": []})
            ).background_tasks
        )
        if model_endpoints_creation_background_tasks:
            self._check_model_endpoint_task_state(
                db=db,
                background_task=model_endpoints_creation_background_tasks[0],
                wait_for_completion=False,
            )

        return self._enrich_command_from_status()

    def _enrich_command_from_status(self):
        # NOTE: on older mlrun versions & nuclio versions, function are exposed via NodePort
        #       now, functions can be not exposed (using service type ClusterIP) and hence
        #       for BC we first try to populate the external invocation url, and then
        #       if not exists, take the internal invocation url
        if (
            self.status.external_invocation_urls
            and self.status.external_invocation_urls[0] != ""
        ):
            self.spec.command = f"http://{self.status.external_invocation_urls[0]}"
        elif (
            self.status.internal_invocation_urls
            and self.status.internal_invocation_urls[0] != ""
        ):
            self.spec.command = f"http://{self.status.internal_invocation_urls[0]}"
        elif self.status.address and self.status.address != "":
            self.spec.command = f"http://{self.status.address}"

        logger.info(
            "Successfully deployed function",
            internal_invocation_urls=self.status.internal_invocation_urls,
            external_invocation_urls=self.status.external_invocation_urls,
        )

        self.save(versioned=False)

        return self.spec.command

    def _wait_for_function_deployment(self, db, verbose=False):
        state = ""
        last_log_timestamp = 1
        while state not in ["ready", "error", "unhealthy"]:
            sleep(
                int(mlrun.mlconf.httpdb.logs.nuclio.pull_deploy_status_default_interval)
            )
            try:
                text, last_log_timestamp = db.get_nuclio_deploy_status(
                    self, last_log_timestamp=last_log_timestamp, verbose=verbose
                )
            except mlrun.db.RunDBError:
                raise ValueError("function or deploy process not found")
            state = self.status.state
            if text:
                print(text)

        if state != "ready":
            logger.error("Nuclio function failed to deploy", function_state=state)
            raise RunError(f"Function {self.metadata.name} deployment failed")

    @min_nuclio_versions("1.5.20", "1.6.10")
    def with_node_selection(
        self,
        node_name: typing.Optional[str] = None,
        node_selector: typing.Optional[dict[str, str]] = None,
        affinity: typing.Optional[client.V1Affinity] = None,
        tolerations: typing.Optional[list[client.V1Toleration]] = None,
    ):
        """k8s node selection attributes"""
        if tolerations and not validate_nuclio_version_compatibility("1.7.5"):
            raise mlrun.errors.MLRunIncompatibleVersionError(
                "tolerations are only supported from nuclio version 1.7.5"
            )
        super().with_node_selection(node_name, node_selector, affinity, tolerations)

    @min_nuclio_versions("1.8.6")
    def with_preemption_mode(self, mode):
        """
        Preemption mode controls whether pods can be scheduled on preemptible nodes.
        Tolerations, node selector, and affinity are populated on preemptible nodes corresponding to the function spec.

        The supported modes are:

        * **allow** - The function can be scheduled on preemptible nodes
        * **constrain** - The function can only run on preemptible nodes
        * **prevent** - The function cannot be scheduled on preemptible nodes
        * **none** - No preemptible configuration will be applied on the function

        The default preemption mode is configurable in mlrun.mlconf.function_defaults.preemption_mode,
        by default it's set to **prevent**

        :param mode: allow | constrain | prevent | none defined in :py:class:`~mlrun.common.schemas.PreemptionModes`
        """
        super().with_preemption_mode(mode=mode)

    @min_nuclio_versions("1.6.18")
    def with_priority_class(self, name: typing.Optional[str] = None):
        """k8s priority class"""
        super().with_priority_class(name)

    def with_service_type(
        self,
        service_type: str,
        add_templated_ingress_host_mode: typing.Optional[str] = None,
    ):
        """
        Enables to control the service type of the pod and the addition of templated ingress host

        :param service_type:                      service type (ClusterIP, NodePort), defaults to
                                                  mlrun.mlconf.httpdb.nuclio.service_type
        :param add_templated_ingress_host_mode:   add templated ingress host mode (never, always, onClusterIP),
                                                  see mlrun.mlconf.httpdb.nuclio.add_templated_ingress_host_mode
                                                  for the default and more information

        """
        self.spec.service_type = service_type
        self.spec.add_templated_ingress_host_mode = add_templated_ingress_host_mode

    def set_state_thresholds(
        self,
        state_thresholds: dict[str, int],
        patch: bool = True,
    ):
        raise NotImplementedError(
            "State thresholds do not apply for nuclio as it has its own function pods healthiness monitoring"
        )

    @min_nuclio_versions("1.13.1")
    def disable_default_http_trigger(
        self,
    ):
        """
        Disables nuclio's default http trigger creation
        """
        self.spec.disable_default_http_trigger = True

    @min_nuclio_versions("1.13.1")
    def enable_default_http_trigger(
        self,
    ):
        """
        Enables nuclio's default http trigger creation
        """
        self.spec.disable_default_http_trigger = False

    def skip_image_enrichment(self):
        # make sure the API does not enrich the base image if the function is not a python function
        return self.spec.nuclio_runtime and "python" not in self.spec.nuclio_runtime

    def _get_state(
        self,
        last_log_timestamp=0,
        verbose=False,
        raise_on_exception=True,
    ) -> tuple[str, str, typing.Optional[float]]:
        try:
            text, last_log_timestamp = self._get_db().get_nuclio_deploy_status(
                self, last_log_timestamp=last_log_timestamp, verbose=verbose
            )
        except mlrun.db.RunDBError:
            if raise_on_exception:
                return "", "", None
            raise ValueError("function or deploy process not found")
        return self.status.state, text, last_log_timestamp

    def _get_runtime_env(self):
        # for runtime specific env var enrichment (before deploy)
        runtime_env = {
            mlrun.common.constants.MLRUN_ACTIVE_PROJECT: self.metadata.project
            or mlconf.active_project,
        }
        if mlconf.httpdb.api_url:
            runtime_env["MLRUN_DBPATH"] = mlconf.httpdb.api_url
        if mlconf.namespace:
            runtime_env["MLRUN_NAMESPACE"] = mlconf.namespace
        if self.metadata.credentials.access_key:
            runtime_env[
                mlrun.common.runtimes.constants.FunctionEnvironmentVariables.auth_session
            ] = self.metadata.credentials.access_key
        return runtime_env

    def _get_serving_spec(self):
        return None

    def _get_nuclio_config_spec_env(self):
        env_dict = {}
        external_source_env_dict = {}

        api = client.ApiClient()
        for env_var in self.spec.env:
            # sanitize env if not sanitized
            if isinstance(env_var, dict):
                sanitized_env_var = env_var
            else:
                sanitized_env_var = api.sanitize_for_serialization(env_var)

            value = sanitized_env_var.get("value")
            if value is not None:
                env_dict[sanitized_env_var.get("name")] = value
                continue

            value_from = sanitized_env_var.get("valueFrom")
            if value_from is not None:
                external_source_env_dict[sanitized_env_var.get("name")] = value_from

        for key, value in self._get_runtime_env().items():
            env_dict[key] = value

        return env_dict, external_source_env_dict

    def deploy_step(
        self,
        project="",
        models=None,
        env=None,
        tag=None,
        verbose=None,
        use_function_from_db=None,
    ):
        """return as a Kubeflow pipeline step (ContainerOp), recommended to use mlrun.deploy_function() instead

        :param project:        project name, defaults to function project
        :param models:         model name and paths
        :param env:            dict of environment variables
        :param tag:            version tag
        :param verbose:        verbose output
        :param use_function_from_db:  use the function from the DB instead of the local function object
        """
        models = {} if models is None else models
        function_name = self.metadata.name or "function"
        name = f"deploy_{function_name}"
        project = project or self.metadata.project
        if models and isinstance(models, dict):
            models = [{"key": k, "model_path": v} for k, v in models.items()]

        # verify auto mount is applied (with the client credentials)
        self.try_auto_mount_based_on_config()

        if (
            use_function_from_db
            or use_function_from_db is None
            # if the function contain KFP PipelineParams (futures) pass the full spec to the
            # ContainerOp this way KFP will substitute the params with previous step outputs
            and not self._has_pipeline_param()
        ):
            url = self.save(versioned=True, refresh=True)
        else:
            url = None

        return deploy_op(
            name,
            self,
            func_url=url,
            project=project,
            models=models,
            env=env,
            tag=tag,
            verbose=verbose,
        )

    def invoke(
        self,
        path: str,
        body: typing.Optional[typing.Union[str, bytes, dict]] = None,
        method: typing.Optional[str] = None,
        headers: typing.Optional[dict] = None,
        force_external_address: bool = False,
        auth_info: AuthInfo = None,
        mock: typing.Optional[bool] = None,
        **http_client_kwargs,
    ):
        """Invoke the remote (live) function and return the results

        example::

            function.invoke("/api", body={"inputs": x})

        :param path:     request sub path (e.g. /images)
        :param body:     request body (str, bytes or a dict for json requests)
        :param method:   HTTP method (GET, PUT, ..)
        :param headers:  key/value dict with http headers
        :param force_external_address:   use the external ingress URL
        :param auth_info: service AuthInfo
        :param mock:     use mock server vs a real Nuclio function (for local simulations)
        :param http_client_kwargs:   allow the user to pass any parameter supported in requests.request method
                                     see this link for more information:
                                     https://requests.readthedocs.io/en/latest/api/#requests.request
        """
        if not method:
            method = "POST" if body else "GET"

        # if no path was provided, use the default handler to be invoked
        if not path and self.spec.default_handler:
            path = self.spec.default_handler

        if (self._mock_server and mock is None) or mlconf.use_nuclio_mock(mock):
            # if we deployed mock server or in simulated nuclio environment use mock
            if not self._mock_server:
                self._set_as_mock(True)
            return self._mock_server.test(path, body, method, headers)

        # clear the mock server when using the real endpoint
        self._mock_server = None

        if "://" not in path:
            path = self._resolve_invocation_url(path, force_external_address)

        if headers is None:
            headers = {}

        # if function is scaled to zero, let the DLX know we want to wake it up
        full_function_name = get_fullname(
            self.metadata.name, self.metadata.project, self.metadata.tag
        )
        headers.setdefault("x-nuclio-target", full_function_name)
        if not http_client_kwargs:
            http_client_kwargs = {}
        if body:
            if isinstance(body, (str, bytes)):
                http_client_kwargs["data"] = body
            else:
                http_client_kwargs["json"] = body
        try:
            logger.debug("Invoking function", method=method, path=path)
            if not getattr(self, "_http_session", None):
                self._http_session = requests.Session()
            resp = self._http_session.request(
                method, path, headers=headers, **http_client_kwargs
            )
        except OSError as err:
            raise OSError(
                f"error: cannot run function at url {path}, {err_to_str(err)}"
            )
        if not resp.ok:
            raise RuntimeError(f"bad function response {resp.status_code}: {resp.text}")

        data = resp.content
        if resp.headers["content-type"] == "application/json":
            data = json.loads(data)
        return data

    def with_sidecar(
        self,
        name: typing.Optional[str] = None,
        image: typing.Optional[str] = None,
        ports: typing.Optional[typing.Union[int, list[int]]] = None,
        command: typing.Optional[str] = None,
        args: typing.Optional[list[str]] = None,
    ):
        """
        Add a sidecar container to the function pod

        :param name:    Sidecar container name.
        :param image:   Sidecar container image.
        :param ports:   Sidecar container ports to expose. Can be a single port or a list of ports.
        :param command: Sidecar container command instead of the image entrypoint.
        :param args:    Sidecar container command args (requires command to be set).
        """
        name = name or f"{self.metadata.name}-sidecar"
        sidecar = self._set_sidecar(name)
        if image:
            sidecar["image"] = image

        ports = mlrun.utils.helpers.as_list(ports)
        if len(ports) > 1:
            mlrun.runtimes.nuclio.multiple_port_sidecar_is_supported()

        # according to RFC-6335, port name should be less than 15 characters,
        # so we truncate it if needed and leave room for the index
        port_name = name[:13].rstrip("-_") if len(name) > 13 else name
        sidecar["ports"] = [
            {
                "name": f"{port_name}-{i}",
                "containerPort": port,
                "protocol": "TCP",
            }
            for i, port in enumerate(ports)
        ]

        # if it is a redeploy, 'command' might be set with the previous invocation url.
        # in this case, we don't want to use it as the sidecar command
        if command and not command.startswith("http"):
            sidecar["command"] = mlrun.utils.helpers.as_list(command)

        if args and sidecar.get("command"):
            sidecar["args"] = mlrun.utils.helpers.as_list(args)

        # put the configured resources on the sidecar container instead of the reverse proxy container
        if self.spec.resources:
            sidecar["resources"] = self.spec.resources
            self.spec.resources = None

    def _set_sidecar(self, name: str) -> dict:
        self.spec.config.setdefault("spec.sidecars", [])
        sidecars = self.spec.config["spec.sidecars"]
        for sidecar in sidecars:
            if sidecar["name"] == name:
                return sidecar

        sidecars.append({"name": name})
        return sidecars[-1]

    def _trigger_of_kind_exists(self, kind: str) -> bool:
        if not self.spec.config:
            return False

        for name, spec in self.spec.config.items():
            if isinstance(spec, dict):
                if spec.get("kind") == kind:
                    return True

        return False

    def _pre_run_validations(self):
        if self.spec.function_kind != "mlrun":
            raise RunError(
                '.run() can only be execute on "mlrun" kind'
                ', recreate with function kind "mlrun"'
            )

        state = self.status.state
        if state != "ready":
            if state:
                raise RunError(f"Cannot run, function in state {state}")
            state, _, _ = self._get_state(raise_on_exception=True)
            if state != "ready":
                logger.info("Starting nuclio build!")
                self.deploy()

    def _run(self, runobj: RunObject, execution):
        self._pre_run_validations()
        self.store_run(runobj)
        if self._secrets:
            runobj.spec.secret_sources = self._secrets.to_serial()
        log_level = execution.log_level
        command = self.spec.command
        if runobj.spec.handler:
            command = f"{command}/{runobj.spec.handler_name}"
        headers = {"x-nuclio-log-level": log_level}
        try:
            resp = requests.put(command, json=runobj.to_dict(), headers=headers)
        except OSError as err:
            logger.error(f"error invoking function: {err_to_str(err)}")
            raise OSError(f"error: cannot run function at url {command}")

        if not resp.ok:
            logger.error(f"bad function resp!!\n{resp.text}")
            raise RunError("bad function response")

        logs = resp.headers.get("X-Nuclio-Logs")
        if logs:
            log_std(self._db_conn, runobj, parse_logs(logs))

        return self._update_state(resp.json())

    def _run_many(self, generator, execution, runobj: RunObject):
        self._pre_run_validations()
        tasks = generator.generate(runobj)
        secrets = self._secrets.to_serial() if self._secrets else None
        log_level = execution.log_level
        headers = {"x-nuclio-log-level": log_level}

        command = self.spec.command
        if runobj.spec.handler:
            command = f"{command}/{runobj.spec.handler_name}"
        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(
            self._invoke_async(tasks, command, headers, secrets, generator=generator)
        )

        loop.run_until_complete(future)
        return future.result()

    def _update_state(self, rundict: dict):
        last_state = get_in(rundict, "status.state", "")
        if last_state != "error":
            update_in(rundict, "status.state", "completed")
        self._store_run_dict(rundict)
        return rundict

    async def _invoke_async(self, tasks, url, headers, secrets, generator):
        results = RunList()
        runs = []
        num_errors = 0
        stop = False
        parallel_runs = generator.options.parallel_runs or 1
        semaphore = asyncio.Semaphore(parallel_runs)

        async with ClientSession() as session:
            for task in tasks:
                # TODO: store run using async calls to improve performance
                self.store_run(task)
                task.spec.secret_sources = secrets or []
                resp = submit(session, url, task, semaphore, headers=headers)
                runs.append(
                    asyncio.ensure_future(
                        resp,
                    )
                )

            for result in asyncio.as_completed(runs):
                status, resp, logs, task = await result

                if status != 200:
                    err_message = f"failed to access {url} - {resp}"
                    # TODO: store logs using async calls to improve performance
                    log_std(
                        self._db_conn,
                        task,
                        parse_logs(logs) if logs else None,
                        err_message,
                        silent=True,
                    )
                    # TODO: update run using async calls to improve performance
                    results.append(self._update_run_state(task=task, err=err_message))
                    num_errors += 1
                else:
                    if logs:
                        log_std(self._db_conn, task, parse_logs(logs))
                    resp = self._update_run_state(json.loads(resp))
                    state = get_in(resp, "status.state", "")
                    if state == "error":
                        num_errors += 1
                    results.append(resp)

                    run_results = get_in(resp, "status.results", {})
                    stop = generator.eval_stop_condition(run_results)
                    if stop:
                        logger.info(
                            f"Reached early stop condition ({generator.options.stop_condition}), stopping iterations!"
                        )
                        break

                if num_errors > generator.max_errors:
                    logger.error("Max errors reached, stopping iterations!")
                    stop = True
                    break

        if stop:
            for task in runs:
                task.cancel()
        return results

    def _resolve_invocation_url(self, path, force_external_address):
        # internal / external invocation urls is a nuclio >= 1.6.x feature
        # try to infer the invocation url from the internal and if not exists, use external.
        # $$$$ we do not want to use the external invocation url (e.g.: ingress, nodePort, etc.)

        # if none of urls is set, function was deployed with watch=False
        # and status wasn't fetched with Nuclio
        # _get_state fetches the state and updates url
        if (
            not self.status.address
            and not self.status.internal_invocation_urls
            and not self.status.external_invocation_urls
        ):
            state, _, _ = self._get_state()
            if state not in ["ready", "scaledToZero"]:
                logger.warning(f"Function is in the {state} state")

        # prefer internal invocation url if running inside k8s cluster
        if (
            not force_external_address
            and self.status.internal_invocation_urls
            and mlrun.k8s_utils.is_running_inside_kubernetes_cluster()
        ):
            url = mlrun.utils.helpers.join_urls(
                f"http://{self.status.internal_invocation_urls[0]}", path
            )
            logger.debug(
                f"Using internal invocation url {url}. Make sure you have network access to the k8s cluster. "
                f"Otherwise, set force_external_address to True"
            )
            return url

        if self.status.external_invocation_urls:
            return mlrun.utils.helpers.join_urls(
                f"http://{self.status.external_invocation_urls[0]}", path
            )

        if not self.status.address:
            # if there is no address
            # here we check that if default http trigger is disabled, function contains a custom http trigger
            # Otherwise, the function is not invokable, so we raise an error
            if (
                not self._trigger_of_kind_exists(kind="http")
                and self.spec.disable_default_http_trigger
            ):
                raise mlrun.errors.MLRunPreconditionFailedError(
                    "Default http trigger creation is disabled and there is no any other custom http trigger, "
                    "so function can not be invoked via http. Either enable default http trigger creation or "
                    "create custom http trigger"
                )
            else:
                raise ValueError("no function address first run .deploy()")
        else:
            return mlrun.utils.helpers.join_urls(f"http://{self.status.address}", path)

    def _update_credentials_from_remote_build(self, remote_data):
        self.metadata.credentials = remote_data.get("metadata", {}).get(
            "credentials", {}
        )

        credentials_env_var_names = ["V3IO_ACCESS_KEY", "MLRUN_AUTH_SESSION"]
        new_env = []

        # the env vars in the local spec and remote spec are in the format of a list of dicts
        # e.g.:
        # env = [
        #   {
        #     "name": "V3IO_ACCESS_KEY",
        #     "value": "some-value"
        #   },
        #   ...
        # ]
        # remove existing credentials env vars
        for env in self.spec.env:
            if isinstance(env, dict):
                env_name = env["name"]
            elif isinstance(env, client.V1EnvVar):
                env_name = env.name
            else:
                continue

            if env_name not in credentials_env_var_names:
                new_env.append(env)

        # add credentials env vars from remote build
        for remote_env in remote_data.get("spec", {}).get("env", []):
            if remote_env.get("name") in credentials_env_var_names:
                new_env.append(remote_env)

        # update nuclio-specific credentials
        self.mask_sensitive_data_in_config()

        self.spec.env = new_env

    def _set_as_mock(self, enable):
        # todo: create mock_server for Nuclio
        if enable:
            raise NotImplementedError(
                "Mock (simulation) is currently not supported for Nuclio, Turn off the mock (mock=False) "
                "and make sure Nuclio is installed for real deployment to Nuclio"
            )

    def get_url(
        self,
        force_external_address: bool = False,
        # leaving auth_info for BC
        # TODO: remove in 1.12.0
        auth_info: AuthInfo = None,
    ):
        """
        This method returns function's url.

        :param force_external_address:   use the external ingress URL
        :param auth_info:                service AuthInfo

        :return: returns function's url
        """
        if auth_info:
            warnings.warn(
                "'auth_info' is deprecated in 1.10.0 and will be removed in 1.12.0.",
                # TODO: Remove this in 1.12.0
                FutureWarning,
            )
        return self._resolve_invocation_url("", force_external_address)

    @staticmethod
    def _resolve_v3io_access_key():
        # Nuclio supports generating access key for v3io stream trigger only from version 1.13.11
        if validate_nuclio_version_compatibility("1.13.11"):
            return mlrun.model.Credentials.generate_access_key
        return None

    def _check_model_endpoint_task_state(
        self,
        db: mlrun.db.RunDBInterface,
        background_task: mlrun.common.schemas.BackgroundTask,
        wait_for_completion: bool,
    ):
        if wait_for_completion:
            background_task = db._wait_for_background_task_to_reach_terminal_state(
                name=background_task.metadata.name, project=self.metadata.project
            )
        else:
            background_task = db.get_project_background_task(
                project=self.metadata.project, name=background_task.metadata.name
            )
        if (
            background_task.status.state
            in mlrun.common.schemas.BackgroundTaskState.terminal_states()
        ):
            logger.info(
                f"Model endpoint creation task completed with state {background_task.status.state}"
            )
        else:
            logger.warning(
                f"Model endpoint creation task is still running with state {background_task.status.state}"
                f"You can use the serving function, but it won't be monitored for the next few minutes"
            )


def parse_logs(logs):
    logs = json.loads(logs)
    lines = ""
    for line in logs:
        extra = []
        for key, value in line.items():
            if key not in ["time", "level", "name", "message"]:
                extra.append(f"{key}={value}")
        extra = ", ".join(extra)
        time = datetime.fromtimestamp(float(line["time"]) / 1000).strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )
        lines += f"{time}  {line['level']:<6} {line['message']}  {extra}\n"

    return lines


async def submit(session, url, run, semaphore, headers=None):
    async with semaphore:
        async with session.put(url, json=run.to_dict(), headers=headers) as response:
            text = await response.text()
            logs = response.headers.get("X-Nuclio-Logs", None)
            return response.status, text, logs, run


def fake_nuclio_context(body, headers=None):
    return nuclio.Context(), nuclio.Event(body=body, headers=headers)


def get_fullname(name, project, tag):
    if project:
        name = f"{project}-{name}"
    if tag and tag != "latest":
        name = f"{name}-{tag}"
    return name


def get_nuclio_deploy_status(
    name,
    project,
    tag,
    last_log_timestamp=0,
    verbose=False,
    resolve_address=True,
    auth_info: AuthInfo = None,
):
    """
    Get nuclio function deploy status

    :param name:                function name
    :param project:             project name
    :param tag:                 function tag
    :param last_log_timestamp:  last log timestamp
    :param verbose:             print logs
    :param resolve_address:     whether to resolve function address
    :param auth_info:           authentication information
    """
    api_address = find_dashboard_url(mlconf.nuclio_dashboard_url)
    name = get_fullname(name, project, tag)
    get_err_message = f"Failed to get function {name} deploy status"

    try:
        (
            state,
            address,
            last_log_timestamp,
            outputs,
            function_status,
        ) = get_deploy_status(
            api_address,
            name,
            last_log_timestamp,
            verbose,
            resolve_address,
            return_function_status=True,
            auth_info=auth_info.to_nuclio_auth_info() if auth_info else None,
        )
    except requests.exceptions.ConnectionError as exc:
        mlrun.errors.raise_for_status(
            exc.response,
            get_err_message,
        )

    except nuclio.utils.DeployError as exc:
        if exc.err:
            mlrun.errors.raise_for_status(
                exc.err.response,
                get_err_message,
            )
        raise exc
    else:
        text = "\n".join(outputs) if outputs else ""
        return state, address, name, last_log_timestamp, text, function_status


def enrich_nuclio_function_from_headers(
    func: RemoteRuntime,
    headers: dict,
):
    func.status.state = headers.get("x-mlrun-function-status", "")
    func.status.address = headers.get("x-mlrun-address", "")
    func.status.nuclio_name = headers.get("x-mlrun-name", "")
    func.status.internal_invocation_urls = (
        headers.get("x-mlrun-internal-invocation-urls", "").split(",")
        if headers.get("x-mlrun-internal-invocation-urls")
        else []
    )
    func.status.external_invocation_urls = (
        headers.get("x-mlrun-external-invocation-urls", "").split(",")
        if headers.get("x-mlrun-external-invocation-urls")
        else []
    )
    func.status.container_image = headers.get("x-mlrun-container-image", "")


@min_nuclio_versions("1.14.14")
def multiple_port_sidecar_is_supported():
    # multiple ports are supported from nuclio version 1.14.14
    # this method exists only for running the min_nuclio_versions decorator
    return True



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Optional, Union

import mlrun.alerts
import mlrun.common.formatters
import mlrun.common.runtimes.constants
import mlrun.common.schemas
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.errors
import mlrun.lists
import mlrun.model_monitoring

from ..config import config
from ..utils import logger
from .base import RunDBInterface


class NopDB(RunDBInterface):
    def __init__(self, url=None, *args, **kwargs):
        self.url = url

    def __getattribute__(self, attr):
        def nop(*args, **kwargs):
            env_var_message = (
                "MLRUN_DBPATH is misconfigured. Set this environment variable to the URL of the API "
                "server in order to connect"
            )
            if config.httpdb.nop_db.raise_error:
                raise mlrun.errors.MLRunBadRequestError(env_var_message)

            if config.httpdb.nop_db.verbose:
                logger.warning(
                    "Could not detect path to API server, not connected to API server!"
                )
                logger.warning(env_var_message)

            return

        # ignore __class__ because __getattribute__ overrides the parent class's method and it spams logs
        if attr in ["connect", "__class__"]:
            return super().__getattribute__(attr)
        else:
            nop()
            return super().__getattribute__(attr)

    def connect(self, secrets=None):
        pass

    def store_log(self, uid, project="", body=None, append=False):
        pass

    def get_log(self, uid, project="", offset=0, size=0, attempt=None):
        pass

    def store_run(self, struct, uid, project="", iter=0):
        pass

    def update_run(self, updates: dict, uid, project="", iter=0):
        pass

    def set_run_retrying_status(
        self, project: str, name: str, run_id: str, retrying: bool
    ):
        pass

    def abort_run(self, uid, project="", iter=0, timeout=45, status_text=""):
        pass

    def push_run_notifications(
        self,
        uid,
        project="",
        timeout=45,
    ):
        pass

    def refresh_smtp_configuration(self):
        pass

    def push_pipeline_notifications(
        self,
        pipeline_id,
        project="",
        notifications=None,
        timeout=45,
    ):
        pass

    def list_runtime_resources(
        self,
        project: Optional[str] = None,
        label_selector: Optional[str] = None,
        kind: Optional[str] = None,
        object_id: Optional[str] = None,
        group_by: Optional[
            mlrun.common.schemas.ListRuntimeResourcesGroupByField
        ] = None,
    ) -> Union[
        mlrun.common.schemas.RuntimeResourcesOutput,
        mlrun.common.schemas.GroupedByJobRuntimeResourcesOutput,
        mlrun.common.schemas.GroupedByProjectRuntimeResourcesOutput,
    ]:
        return []

    def read_run(
        self,
        uid,
        project="",
        iter=0,
        format_: mlrun.common.formatters.RunFormat = mlrun.common.formatters.RunFormat.full,
    ):
        pass

    def list_runs(
        self,
        name: Optional[str] = None,
        uid: Optional[Union[str, list[str]]] = None,
        project: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        states: Optional[list[mlrun.common.runtimes.constants.RunStates]] = None,
        sort: bool = True,
        iter: bool = False,
        start_time_from: Optional[datetime.datetime] = None,
        start_time_to: Optional[datetime.datetime] = None,
        last_update_time_from: Optional[datetime.datetime] = None,
        last_update_time_to: Optional[datetime.datetime] = None,
        end_time_from: Optional[datetime.datetime] = None,
        end_time_to: Optional[datetime.datetime] = None,
        partition_by: Union[mlrun.common.schemas.RunPartitionByField, str] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Union[mlrun.common.schemas.SortField, str] = None,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
        max_partitions: int = 0,
        with_notifications: bool = False,
    ):
        return mlrun.lists.RunList()

    def paginated_list_runs(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        return mlrun.lists.RunList(), None

    def del_run(self, uid, project="", iter=0):
        pass

    def del_runs(
        self,
        name: str = "",
        project: str = "",
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        state: Optional[mlrun.common.runtimes.constants.RunStates] = None,
        days_ago: int = 0,
    ):
        pass

    def store_artifact(
        self,
        key,
        artifact,
        iter=None,
        tag="",
        project="",
        tree=None,
    ):
        pass

    def read_artifact(
        self,
        key,
        tag="",
        iter=None,
        project="",
        tree=None,
        uid=None,
        format_: mlrun.common.formatters.ArtifactFormat = mlrun.common.formatters.ArtifactFormat.full,
    ):
        pass

    def list_artifacts(
        self,
        name="",
        project="",
        tag="",
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        since=None,
        until=None,
        iter: Optional[int] = None,
        best_iteration: bool = False,
        kind: Optional[str] = None,
        category: Union[str, mlrun.common.schemas.ArtifactCategories] = None,
        tree: Optional[str] = None,
        parent: Optional[str] = None,
        format_: mlrun.common.formatters.ArtifactFormat = mlrun.common.formatters.ArtifactFormat.full,
        limit: Optional[int] = None,
        partition_by: Optional[
            Union[mlrun.common.schemas.ArtifactPartitionByField, str]
        ] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Optional[
            Union[mlrun.common.schemas.SortField, str]
        ] = mlrun.common.schemas.SortField.updated,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
    ):
        return mlrun.lists.ArtifactList()

    def paginated_list_artifacts(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        return mlrun.lists.ArtifactList(), None

    def del_artifact(
        self,
        key,
        tag="",
        project="",
        tree=None,
        uid=None,
        deletion_strategy: mlrun.common.schemas.artifact.ArtifactsDeletionStrategies = (
            mlrun.common.schemas.artifact.ArtifactsDeletionStrategies.metadata_only
        ),
        secrets: Optional[dict] = None,
        iter=None,
    ):
        pass

    def del_artifacts(
        self,
        name="",
        project="",
        tag="",
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ):
        pass

    def store_function(self, function, name, project="", tag="", versioned=False):
        pass

    def get_function(self, name, project="", tag="", hash_key=""):
        pass

    def delete_function(self, name: str, project: str = ""):
        pass

    def list_functions(
        self,
        name: Optional[str] = None,
        project: Optional[str] = None,
        tag: Optional[str] = None,
        kind: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        states: Optional[list[mlrun.common.schemas.FunctionState]] = None,
        format_: mlrun.common.formatters.FunctionFormat = mlrun.common.formatters.FunctionFormat.full,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
    ):
        return []

    def paginated_list_functions(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        return [], None

    def tag_objects(
        self,
        project: str,
        tag_name: str,
        tag_objects: mlrun.common.schemas.TagObjects,
        replace: bool = False,
    ):
        pass

    def delete_objects_tag(
        self, project: str, tag_name: str, tag_objects: mlrun.common.schemas.TagObjects
    ):
        pass

    def tag_artifacts(
        self, artifacts, project: str, tag_name: str, replace: bool = False
    ):
        pass

    def delete_artifacts_tags(self, artifacts, project: str, tag_name: str):
        pass

    def delete_project(
        self,
        name: str,
        deletion_strategy: mlrun.common.schemas.DeletionStrategy = mlrun.common.schemas.DeletionStrategy.default(),
    ):
        pass

    def store_project(
        self, name: str, project: mlrun.common.schemas.Project
    ) -> mlrun.common.schemas.Project:
        pass

    def patch_project(
        self,
        name: str,
        project: dict,
        patch_mode: mlrun.common.schemas.PatchMode = mlrun.common.schemas.PatchMode.replace,
    ) -> mlrun.common.schemas.Project:
        pass

    def create_project(
        self, project: mlrun.common.schemas.Project
    ) -> mlrun.common.schemas.Project:
        pass

    def list_projects(
        self,
        owner: Optional[str] = None,
        format_: mlrun.common.formatters.ProjectFormat = mlrun.common.formatters.ProjectFormat.name_only,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        state: mlrun.common.schemas.ProjectState = None,
    ) -> mlrun.common.schemas.ProjectsOutput:
        pass

    def get_project(self, name: str) -> mlrun.common.schemas.Project:
        pass

    def list_artifact_tags(
        self,
        project=None,
        category: Union[str, mlrun.common.schemas.ArtifactCategories] = None,
    ):
        pass

    def create_feature_set(
        self,
        feature_set: Union[dict, mlrun.common.schemas.FeatureSet],
        project="",
        versioned=True,
    ) -> dict:
        pass

    def get_feature_set(
        self,
        name: str,
        project: str = "",
        tag: Optional[str] = None,
        uid: Optional[str] = None,
    ) -> dict:
        pass

    def list_features_v2(
        self,
        project: str,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        entities: Optional[list[str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ) -> mlrun.common.schemas.FeaturesOutputV2:
        pass

    def list_entities_v2(
        self,
        project: str,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
    ) -> mlrun.common.schemas.EntitiesOutputV2:
        pass

    def list_feature_sets(
        self,
        project: str = "",
        name: Optional[str] = None,
        tag: Optional[str] = None,
        state: Optional[str] = None,
        entities: Optional[list[str]] = None,
        features: Optional[list[str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        partition_by: Union[
            mlrun.common.schemas.FeatureStorePartitionByField, str
        ] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Union[mlrun.common.schemas.SortField, str] = None,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
        format_: Union[
            str, mlrun.common.formatters.FeatureSetFormat
        ] = mlrun.common.formatters.FeatureSetFormat.full,
    ) -> list[dict]:
        pass

    def store_feature_set(
        self,
        feature_set: Union[dict, mlrun.common.schemas.FeatureSet],
        name=None,
        project="",
        tag=None,
        uid=None,
        versioned=True,
    ):
        pass

    def patch_feature_set(
        self,
        name,
        feature_set: dict,
        project="",
        tag=None,
        uid=None,
        patch_mode: Union[
            str, mlrun.common.schemas.PatchMode
        ] = mlrun.common.schemas.PatchMode.replace,
    ):
        pass

    def delete_feature_set(self, name, project="", tag=None, uid=None):
        pass

    def create_feature_vector(
        self,
        feature_vector: Union[dict, mlrun.common.schemas.FeatureVector],
        project="",
        versioned=True,
    ) -> dict:
        pass

    def get_feature_vector(
        self,
        name: str,
        project: str = "",
        tag: Optional[str] = None,
        uid: Optional[str] = None,
    ) -> dict:
        pass

    def list_feature_vectors(
        self,
        project: str = "",
        name: Optional[str] = None,
        tag: Optional[str] = None,
        state: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        partition_by: Union[
            mlrun.common.schemas.FeatureStorePartitionByField, str
        ] = None,
        rows_per_partition: int = 1,
        partition_sort_by: Union[mlrun.common.schemas.SortField, str] = None,
        partition_order: Union[
            mlrun.common.schemas.OrderType, str
        ] = mlrun.common.schemas.OrderType.desc,
    ) -> list[dict]:
        pass

    def store_feature_vector(
        self,
        feature_vector: Union[dict, mlrun.common.schemas.FeatureVector],
        name=None,
        project="",
        tag=None,
        uid=None,
        versioned=True,
    ):
        pass

    def patch_feature_vector(
        self,
        name,
        feature_vector_update: dict,
        project="",
        tag=None,
        uid=None,
        patch_mode: Union[
            str, mlrun.common.schemas.PatchMode
        ] = mlrun.common.schemas.PatchMode.replace,
    ):
        pass

    def delete_feature_vector(self, name, project="", tag=None, uid=None):
        pass

    def get_pipeline(
        self,
        run_id: str,
        namespace: Optional[str] = None,
        timeout: int = 30,
        format_: Union[
            str, mlrun.common.formatters.PipelineFormat
        ] = mlrun.common.formatters.PipelineFormat.summary,
        project: Optional[str] = None,
    ):
        pass

    def retry_pipeline(
        self,
        run_id: str,
        project: str,
        namespace: Optional[str] = None,
        timeout: int = 30,
    ):
        pass

    def list_pipelines(
        self,
        project: str,
        namespace: Optional[str] = None,
        sort_by: str = "",
        page_token: str = "",
        filter_: str = "",
        format_: Union[
            str, mlrun.common.formatters.PipelineFormat
        ] = mlrun.common.formatters.PipelineFormat.metadata_only,
        page_size: Optional[int] = None,
    ) -> mlrun.common.schemas.PipelinesOutput:
        return mlrun.common.schemas.PipelinesOutput(runs=[], total_size=0)

    def create_project_secrets(
        self,
        project: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        secrets: Optional[dict] = None,
    ):
        pass

    def list_project_secrets(
        self,
        project: str,
        token: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        secrets: Optional[list[str]] = None,
    ) -> mlrun.common.schemas.SecretsData:
        pass

    def list_project_secret_keys(
        self,
        project: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        token: Optional[str] = None,
    ) -> mlrun.common.schemas.SecretKeysData:
        pass

    def delete_project_secrets(
        self,
        project: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.kubernetes,
        secrets: Optional[list[str]] = None,
    ):
        pass

    def create_user_secrets(
        self,
        user: str,
        provider: Union[
            str, mlrun.common.schemas.SecretProviderName
        ] = mlrun.common.schemas.SecretProviderName.vault,
        secrets: Optional[dict] = None,
    ):
        pass

    def create_model_endpoint(
        self,
        model_endpoint: mlrun.common.schemas.ModelEndpoint,
        creation_strategy: Optional[
            mm_constants.ModelEndpointCreationStrategy
        ] = mm_constants.ModelEndpointCreationStrategy.INPLACE,
    ) -> mlrun.common.schemas.ModelEndpoint:
        pass

    def delete_model_endpoint(
        self,
        name: str,
        project: str,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        endpoint_id: Optional[str] = None,
    ):
        pass

    def list_model_endpoints(
        self,
        project: str,
        names: Optional[Union[str, list[str]]] = None,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        model_name: Optional[str] = None,
        model_tag: Optional[str] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        tsdb_metrics: bool = False,
        metric_list: Optional[list[str]] = None,
        top_level: bool = False,
        modes: Optional[
            Union[mm_constants.EndpointMode, list[mm_constants.EndpointMode]]
        ] = None,
        uids: Optional[list[str]] = None,
        latest_only: bool = False,
    ) -> mlrun.common.schemas.ModelEndpointList:
        pass

    def get_model_endpoint(
        self,
        name: str,
        project: str,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        endpoint_id: Optional[str] = None,
        tsdb_metrics: bool = True,
        metric_list: Optional[list[str]] = None,
        feature_analysis: bool = False,
    ) -> mlrun.common.schemas.ModelEndpoint:
        pass

    def patch_model_endpoint(
        self,
        name: str,
        project: str,
        attributes: dict,
        function_name: Optional[str] = None,
        function_tag: Optional[str] = None,
        endpoint_id: Optional[str] = None,
    ) -> mlrun.common.schemas.ModelEndpoint:
        pass

    def create_hub_source(
        self, source: Union[dict, mlrun.common.schemas.IndexedHubSource]
    ):
        pass

    def store_hub_source(
        self,
        source_name: str,
        source: Union[dict, mlrun.common.schemas.IndexedHubSource],
    ):
        pass

    def list_hub_sources(
        self,
        item_name: Optional[str] = None,
        tag: Optional[str] = None,
        version: Optional[str] = None,
        item_type: mlrun.common.schemas.hub.HubSourceType = mlrun.common.schemas.hub.HubSourceType.functions,
    ):
        pass

    def get_hub_source(self, source_name: str):
        pass

    def delete_hub_source(self, source_name: str):
        pass

    def get_hub_catalog(
        self,
        source_name: str,
        channel: Optional[str] = None,
        version: Optional[str] = None,
        tag: Optional[str] = None,
        force_refresh: bool = False,
        object_type: mlrun.common.schemas.hub.HubSourceType = mlrun.common.schemas.hub.HubSourceType.functions,
    ):
        pass

    def get_hub_item(
        self,
        source_name: str,
        item_name: str,
        channel: str = "development",
        version: Optional[str] = None,
        tag: str = "latest",
        force_refresh: bool = False,
        item_type: mlrun.common.schemas.hub.HubSourceType = mlrun.common.schemas.hub.HubSourceType.functions,
    ):
        pass

    def get_hub_asset(
        self,
        source_name: str,
        item_name: str,
        asset_name: str,
        version: Optional[str] = None,
        tag: str = "latest",
        item_type: mlrun.common.schemas.hub.HubSourceType = mlrun.common.schemas.hub.HubSourceType.functions,
    ):
        pass

    def store_api_gateway(
        self,
        api_gateway: Union[
            mlrun.common.schemas.APIGateway,
            mlrun.runtimes.nuclio.api_gateway.APIGateway,
        ],
        project: Optional[str] = None,
    ) -> mlrun.common.schemas.APIGateway:
        pass

    def list_api_gateways(self, project=None):
        pass

    def get_api_gateway(self, name, project=None):
        pass

    def delete_api_gateway(self, name, project=None):
        pass

    def verify_authorization(
        self,
        authorization_verification_input: mlrun.common.schemas.AuthorizationVerificationInput,
    ):
        pass

    def remote_builder(
        self,
        func: "mlrun.runtimes.BaseRuntime",
        with_mlrun: bool,
        mlrun_version_specifier: Optional[str] = None,
        skip_deployed: bool = False,
        builder_env: Optional[dict] = None,
        force_build: bool = False,
    ):
        pass

    def deploy_nuclio_function(
        self,
        func: "mlrun.runtimes.RemoteRuntime",
        builder_env: Optional[dict] = None,
    ):
        pass

    def get_builder_status(
        self,
        func: "mlrun.runtimes.BaseRuntime",
        offset: int = 0,
        logs: bool = True,
        last_log_timestamp: float = 0.0,
        verbose: bool = False,
        events_offset: int = 0,
    ):
        pass

    def get_nuclio_deploy_status(
        self,
        func: "mlrun.runtimes.RemoteRuntime",
        last_log_timestamp: float = 0.0,
        verbose: bool = False,
    ):
        pass

    def set_run_notifications(
        self,
        project: str,
        runs: list[mlrun.model.RunObject],
        notifications: list[mlrun.model.Notification],
    ):
        pass

    def store_run_notifications(
        self,
        notification_objects: list[mlrun.model.Notification],
        run_uid: str,
        project: Optional[str] = None,
        mask_params: bool = True,
    ):
        pass

    def store_alert_notifications(
        self,
        session,
        notification_objects: list[mlrun.model.Notification],
        alert_id: str,
        project: str,
        mask_params: bool = True,
    ):
        pass

    def get_log_size(self, uid, project=""):
        pass

    def watch_log(self, uid, project="", watch=True, offset=0):
        pass

    def get_datastore_profile(
        self, name: str, project: str
    ) -> Optional[mlrun.common.schemas.DatastoreProfile]:
        pass

    def delete_datastore_profile(self, name: str, project: str):
        pass

    def list_datastore_profiles(
        self, project: str
    ) -> list[mlrun.common.schemas.DatastoreProfile]:
        pass

    def store_datastore_profile(
        self, profile: mlrun.common.schemas.DatastoreProfile, project: str
    ):
        pass

    def function_status(self, project, name, kind, selector):
        pass

    def start_function(
        self,
        func_url: Optional[str] = None,
        function: "mlrun.runtimes.BaseRuntime" = None,
    ):
        pass

    def submit_workflow(
        self,
        project: str,
        name: str,
        workflow_spec: Union[
            "mlrun.projects.pipelines.WorkflowSpec",
            "mlrun.common.schemas.WorkflowSpec",
            dict,
        ],
        arguments: Optional[dict] = None,
        artifact_path: Optional[str] = None,
        source: Optional[str] = None,
        run_name: Optional[str] = None,
        namespace: Optional[str] = None,
        notifications: Optional[list["mlrun.model.Notification"]] = None,
    ) -> "mlrun.common.schemas.WorkflowResponse":
        pass

    def update_model_monitoring_controller(
        self,
        project: str,
        base_period: int = 10,
        image: str = "mlrun/mlrun",
    ):
        pass

    def enable_model_monitoring(
        self,
        project: str,
        base_period: int = 10,
        image: str = "mlrun/mlrun",
        deploy_histogram_data_drift_app: bool = True,
        fetch_credentials_from_sys_config: bool = False,
    ) -> None:
        pass

    def disable_model_monitoring(
        self,
        project: str,
        delete_resources: bool = True,
        delete_stream_function: bool = False,
        delete_histogram_data_drift_app: bool = True,
        delete_user_applications: bool = False,
        user_application_list: Optional[list[str]] = None,
    ) -> bool:
        pass

    def delete_model_monitoring_function(
        self, project: str, functions: list[str]
    ) -> bool:
        pass

    def set_model_monitoring_credentials(
        self,
        project: str,
        credentials: dict[str, Optional[str]],
        replace_creds: bool,
    ) -> None:
        pass

    def delete_model_monitoring_metrics(
        self,
        project: str,
        application_name: str,
        endpoint_ids: Optional[list[str]] = None,
    ) -> None:
        pass

    def get_monitoring_function_summaries(
        self,
        project: str,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        names: Optional[Union[list[str], str]] = None,
        labels: Optional[Union[str, dict[str, Optional[str]], list[str]]] = None,
        include_stats: bool = False,
        include_infra: bool = True,
    ) -> [mlrun.common.schemas.model_monitoring.FunctionSummary]:
        pass

    def get_monitoring_function_summary(
        self,
        project: str,
        function_name: str,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        include_latest_metrics: bool = False,
    ) -> mlrun.common.schemas.model_monitoring.FunctionSummary:
        pass

    def generate_event(
        self, name: str, event_data: Union[dict, mlrun.common.schemas.Event], project=""
    ):
        pass

    def store_alert_config(
        self,
        alert_name: str,
        alert_data: Union[dict, mlrun.alerts.alert.AlertConfig],
        project="",
        force_reset: bool = False,
    ):
        pass

    def get_alert_config(self, alert_name: str, project=""):
        pass

    def list_alerts_configs(self, project=""):
        pass

    def delete_alert_config(self, alert_name: str, project=""):
        pass

    def reset_alert_config(self, alert_name: str, project=""):
        pass

    def get_alert_template(self, template_name: str):
        pass

    def list_alert_templates(self):
        pass

    def list_alert_activations(
        self,
        project: Optional[str] = None,
        name: Optional[str] = None,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        entity: Optional[str] = None,
        severity: Optional[
            list[Union[mlrun.common.schemas.alert.AlertSeverity, str]]
        ] = None,
        entity_kind: Optional[
            Union[mlrun.common.schemas.alert.EventEntityKind, str]
        ] = None,
        event_kind: Optional[Union[mlrun.common.schemas.alert.EventKind, str]] = None,
    ):
        pass

    def paginated_list_alert_activations(
        self,
        *args,
        page: Optional[int] = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        **kwargs,
    ):
        pass

    def get_alert_activation(
        self,
        project,
        activation_id,
    ) -> mlrun.common.schemas.AlertActivation:
        pass

    def get_project_summary(self, project: str):
        pass

    def get_drift_over_time(
        self,
        project: str,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
    ) -> mlrun.common.schemas.model_monitoring.ModelEndpointDriftValues:
        pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import typing
from typing import Optional

import aiohttp
import aiohttp.http_exceptions
from aiohttp_retry import ExponentialRetry, RequestParams, RetryClient, RetryOptionsBase
from aiohttp_retry.client import _RequestContext

from mlrun.config import config
from mlrun.errors import err_to_str
from mlrun.errors import raise_for_status as ml_raise_for_status
from mlrun.utils.helpers import logger as mlrun_logger

DEFAULT_BLACKLISTED_METHODS = [
    "POST",
    "PUT",
    "PATCH",
]


class AsyncClientWithRetry(RetryClient):
    """
    Extends by adding retry logic on both error statuses and certain exceptions.
    """

    def __init__(
        self,
        max_retries: int = config.http_retry_defaults.max_retries,
        retry_backoff_factor: float = config.http_retry_defaults.backoff_factor,
        retry_on_status_codes: list[int] = config.http_retry_defaults.status_codes,
        retry_on_exception: bool = True,
        raise_for_status: bool = True,
        blacklisted_methods: typing.Optional[list[str]] = None,
        logger: Optional[logging.Logger] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            retry_options=ExponentialRetryOverride(
                retry_on_exception=retry_on_exception,
                blacklisted_methods=blacklisted_methods,
                attempts=max_retries,
                statuses=retry_on_status_codes,
                factor=retry_backoff_factor,
                # do not retry on all service errors. we want to explicit the status codes we want to retry on
                retry_all_server_errors=False,
            ),
            logger=logger or mlrun_logger,
            raise_for_status=raise_for_status,
            **kwargs,
        )

    @property
    def retry_options(self) -> "ExponentialRetryOverride":
        return typing.cast(ExponentialRetryOverride, self._retry_options)

    def methods_blacklist_update_required(self, new_blacklist: str):
        self._retry_options: ExponentialRetryOverride
        return set(self.retry_options.blacklisted_methods).difference(
            set(new_blacklist)
        )

    def _make_requests(
        self,
        params_list: list[RequestParams],
        retry_options: Optional[RetryOptionsBase] = None,
        raise_for_status: Optional[bool] = None,
    ) -> "_CustomRequestContext":
        if retry_options is None:
            retry_options = self._retry_options
        if raise_for_status is None:
            raise_for_status = self._raise_for_status
        return _CustomRequestContext(
            request_func=self._client.request,
            params_list=params_list,
            logger=self._logger,
            retry_options=retry_options,
            raise_for_status=raise_for_status,
        )


class ExponentialRetryOverride(ExponentialRetry):
    # make sure to only add exceptions that are raised early in the request. For example, ConnectionError can be raised
    # during the handling of a request, and therefore should not be retried, as the request might not be idempotent.
    HTTP_RETRYABLE_EXCEPTIONS = [
        # "Connection reset by peer" is raised when the server closes the connection prematurely during TCP handshake.
        ConnectionResetError,
        # "Connection aborted" and "Connection refused" happen when the server doesn't respond at all.
        ConnectionRefusedError,
        ConnectionAbortedError,
        ConnectionError,
        # aiohttp exceptions that can be raised during connection establishment
        aiohttp.ClientConnectionError,
        aiohttp.ServerDisconnectedError,
        asyncio.exceptions.TimeoutError,
    ]

    def __init__(
        self,
        retry_on_exception: typing.Optional[bool] = True,
        blacklisted_methods: typing.Optional[list[str]] = None,
        *args,
        **kwargs,
    ):
        # whether to retry on exceptions
        self.retry_on_exception = retry_on_exception

        # methods that should not be retried
        # do not retry on PUT / PATCH as they might have side effects (not truly idempotent)
        self.blacklisted_methods = (
            blacklisted_methods
            if blacklisted_methods is not None
            else DEFAULT_BLACKLISTED_METHODS
        )

        # default exceptions that should be retried on (when retry_on_exception is True)
        if "exceptions" not in kwargs:
            kwargs["exceptions"] = self.HTTP_RETRYABLE_EXCEPTIONS
        super().__init__(*args, **kwargs)


class _CustomRequestContext(_RequestContext):
    """
    Extends by adding retry logic on both error statuses and certain exceptions.
    """

    @property
    def retry_options(self) -> ExponentialRetryOverride:
        return typing.cast(ExponentialRetryOverride, self._retry_options)

    async def _do_request(self) -> aiohttp.ClientResponse:
        current_attempt = 0
        while True:
            current_attempt += 1
            response = None
            params: typing.Optional[RequestParams] = None
            try:
                try:
                    params = self._params_list[current_attempt - 1]
                except IndexError:
                    params = self._params_list[-1]

                headers = {k: v for k, v in params.headers.items() if v is not None}

                # enrich user agent
                # will help traceability and debugging
                headers[aiohttp.hdrs.USER_AGENT] = (
                    f"{aiohttp.http.SERVER_SOFTWARE} mlrun/{config.version}"
                )

                response: typing.Optional[
                    aiohttp.ClientResponse
                ] = await self._request_func(
                    params.method,
                    params.url,
                    headers=headers,
                    trace_request_ctx={
                        "current_attempt": current_attempt,
                        **(params.trace_request_ctx or {}),
                    },
                    **(params.kwargs or {}),
                )
                retry_wait = self._retry_options.get_timeout(
                    attempt=current_attempt, response=response
                )

                # if the response is not retryable, return it.
                # this is done to prevent the retry logic from running on non-idempotent methods such as POST.
                if not self._is_method_retryable(params.method):
                    self._response = response
                    return response

                # allow user to provide a callback to decide if retry is needed
                if await self._check_response_callback(
                    params, current_attempt, retry_wait, response
                ):
                    self._response = response
                    return response

                last_attempt = current_attempt == self._retry_options.attempts
                if not self._is_status_retryable(response.status) or last_attempt:
                    if self._raise_for_status:
                        ml_raise_for_status(response)

                    self._response = response
                    return response
                else:
                    self._logger.debug(
                        "Request failed on retryable http code, retrying",
                        retry_wait_secs=retry_wait,
                        method=params.method,
                        url=params.url,
                        current_attempt=current_attempt,
                        max_attemps=self._retry_options.attempts,
                        status_code=response.status,
                    )

            except Exception as exc:
                if not self.retry_options.retry_on_exception:
                    self._logger.warning(
                        "Request failed, retry on exception is disabled",
                        method=params.method,
                        url=params.url,
                        exc=err_to_str(exc),
                    )
                    raise exc

                # exhausted all attempts, stop here, return the last response
                exhausted_attempts = current_attempt >= self._retry_options.attempts

                # if the response is not retryable, return now.
                # this is done to prevent the retry logic from running on non-idempotent methods such as POST.
                not_retryable_method = not self._is_method_retryable(params.method)
                is_connection_error = isinstance(
                    exc.__cause__, ConnectionRefusedError
                ) and "[Errno 111] Connect call failed" in str(exc)

                # while method might be blacklisted, we still want to retry on connection errors
                avoid_retry_on_method = not_retryable_method and not is_connection_error
                if exhausted_attempts or avoid_retry_on_method:
                    if response:
                        self._response = response
                        return response
                    raise exc

                # by type
                self.verify_exception_type(exc)

                retry_wait = self._retry_options.get_timeout(
                    attempt=current_attempt, response=None
                )
                self._logger.warning(
                    "Request failed on retryable exception, retrying",
                    retry_wait_secs=retry_wait,
                    method=params.method,
                    url=params.url,
                    current_attempt=current_attempt,
                    max_attemps=self._retry_options.attempts,
                    exc=err_to_str(exc),
                )

            await asyncio.sleep(retry_wait)

    def _is_method_retryable(self, method: str):
        return method not in self.retry_options.blacklisted_methods

    def _is_status_retryable(self, status: int):
        return status in self._retry_options.statuses

    async def _check_response_callback(
        self,
        params: RequestParams,
        retry_count: int,
        retry_wait_secs: float,
        response: aiohttp.ClientResponse,
    ):
        if self._retry_options.evaluate_response_callback is not None:
            try:
                result = self._retry_options.evaluate_response_callback(response)
                if asyncio.iscoroutinefunction(
                    self._retry_options.evaluate_response_callback
                ):
                    return await result
                else:
                    return result
            except Exception as exc:
                self._logger.warning(
                    "Request failed on evaluating response",
                    retry_wait_secs=retry_wait_secs,
                    method=params.method,
                    url=params.url,
                    retry_count=retry_count,
                    max_attemps=self._retry_options.attempts,
                    status_code=response.status,
                    exc=err_to_str(exc),
                )
                return False
        return False

    def verify_exception_type(self, exc):
        for exc_type in self._retry_options.exceptions:
            if isinstance(exc, exc_type):
                return
            if isinstance(exc, aiohttp.ClientConnectorError):
                if isinstance(exc.os_error, exc_type):
                    return
        if exc.__cause__:
            # If the cause exception is retryable, return, otherwise, raise the original exception
            try:
                self.verify_exception_type(exc.__cause__)
            except Exception:
                raise exc
            return
        else:
            raise exc



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import re

from setuptools import setup

import dependencies
import packages

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("mlrun-setup")


def version():
    try:
        with open("mlrun/utils/version/version.json") as version_file:
            version_metadata = json.load(version_file)
            version_ = version_metadata["version"]
            # replace "1.4.0-rc1+rca" with "1.4.0rc1+rca"
            return re.sub(r"(\d+\.\d+\.\d+)-rc(\d+)", r"\1rc\2", version_)
    except (ValueError, KeyError, FileNotFoundError):
        # When installing un-released version (e.g. by doing
        # pip install git+https://github.com/mlrun/mlrun@development)
        # it won't have a version file, so adding some sane default
        logger.warning("Failed resolving version. Ignoring and using 0.0.0+unstable")
        return "0.0.0+unstable"


with open("README.md") as fp:
    long_desc = fp.read()


setup(
    name="mlrun",
    version=version(),
    description="Tracking and config of machine learning runs",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    author="Yaron Haviv",
    author_email="yaronh@iguazio.com",
    license="Apache License 2.0",
    url="https://github.com/mlrun/mlrun",
    packages=packages.packages(
        exclude_packages=[
            "server",
        ]
    ),
    package_data={
        "mlrun": [
            "utils/version/version.json",
            "runtimes/nuclio/application/*.go",
        ]
    },
    keywords=[
        "mlrun",
        "mlops",
        "data-science",
        "machine-learning",
        "experiment-tracking",
    ],
    python_requires=">=3.9, <3.12",
    install_requires=dependencies.base_requirements(),
    extras_require=dependencies.extra_requirements(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
    ],
    zip_safe=False,
    include_package_data=True,
    entry_points={"console_scripts": ["mlrun=mlrun.__main__:main"]},
)



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import tempfile
from collections import defaultdict
from typing import Optional, Union

import mlrun
import mlrun.artifacts.model as model_art
import mlrun.common.schemas
from mlrun.artifacts import Artifact, ArtifactMetadata, ArtifactSpec
from mlrun.utils import StorePrefix, logger

MAX_PROMPT_LENGTH = 1024


class LLMPromptArtifactSpec(ArtifactSpec):
    _dict_fields = ArtifactSpec._dict_fields + [
        "prompt_template",
        "prompt_legend",
        "invocation_config",
        "description",
    ]
    PROMPT_TEMPLATE_KEYS = ("content", "role")
    PROMPT_LEGENDS_KEYS = ("field", "description")

    def __init__(
        self,
        model_artifact: Union[model_art.ModelArtifact, str] = None,
        prompt_template: Optional[list[dict]] = None,
        prompt_path: Optional[str] = None,
        prompt_legend: Optional[dict] = None,
        invocation_config: Optional[dict] = None,
        description: Optional[str] = None,
        target_path: Optional[str] = None,
        **kwargs,
    ):
        if prompt_template and prompt_path:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Cannot specify both 'prompt_template' and 'prompt_path'"
            )
        if prompt_legend:
            self._verify_prompt_legend(prompt_legend)
        if prompt_path:
            self._verify_prompt_path(prompt_path)
        if prompt_template:
            self._verify_prompt_template(prompt_template)
        super().__init__(
            src_path=prompt_path,
            target_path=target_path,
            parent_uri=model_artifact.uri
            if isinstance(model_artifact, model_art.ModelArtifact)
            else model_artifact,
            format=kwargs.pop("format", "") or "json",
            **kwargs,
        )

        self.prompt_template = prompt_template
        self.prompt_legend = prompt_legend
        if invocation_config is not None and not isinstance(invocation_config, dict):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "LLMPromptArtifact invocation_config must be a dictionary or None"
            )
        self.invocation_config = invocation_config or {}
        self.description = description
        self._model_artifact = (
            model_artifact
            if isinstance(model_artifact, model_art.ModelArtifact)
            else None
        )

    def _verify_prompt_template(self, prompt_template):
        if not (
            isinstance(prompt_template, list)
            and all(isinstance(item, dict) for item in prompt_template)
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Expected prompt_template to be a list of dicts"
            )
        for message in prompt_template:
            if set(key.lower() for key in message.keys()) != set(
                self.PROMPT_TEMPLATE_KEYS
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Expected prompt_template to contain dicts with keys "
                    f"{self.PROMPT_TEMPLATE_KEYS}, got {message.keys()}"
                )
            keys_to_pop = []
            for key in message.keys():
                if isinstance(key, str):
                    if not key.islower():
                        message[key.lower()] = message[key]
                        keys_to_pop.append(key)
                else:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"Expected prompt_template to contain dict that only"
                        f" has str keys got {key} of type {type(key)}"
                    )
            for key_to_pop in keys_to_pop:
                message.pop(key_to_pop)

    @property
    def model_uri(self):
        return self.parent_uri

    @staticmethod
    def _verify_prompt_legend(prompt_legend: dict):
        if prompt_legend is None:
            return True
        for place_holder, body_map in prompt_legend.items():
            if isinstance(body_map, dict):
                if body_map.get("field") is None:
                    body_map["field"] = place_holder
                body_map["description"] = body_map.get("description")
                if diff := set(body_map.keys()) - set(
                    LLMPromptArtifactSpec.PROMPT_LEGENDS_KEYS
                ):
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "prompt_legend values must contain only 'field' and "
                        f"'description' keys, got extra fields: {diff}"
                    )
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Wrong prompt_legend format, {place_holder} is not mapped to dict"
                )

    @staticmethod
    def _verify_prompt_path(prompt_path: str):
        with mlrun.datastore.store_manager.object(prompt_path).open(mode="r") as p_file:
            try:
                json.load(p_file)
            except json.JSONDecodeError:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Failed on decoding str in path "
                    f"{prompt_path} expected file to contain a "
                    f"json format."
                )

    def get_body(self):
        if self.prompt_template:
            return json.dumps(self.prompt_template)
        else:
            return None


class LLMPromptArtifact(Artifact):
    """
    LLM Prompt Artifact

    This artifact is used to store and manage LLM prompts.
    Stores the prompt string/path and a link to the related model artifact.
    """

    kind = mlrun.common.schemas.ArtifactCategories.llm_prompt
    _store_prefix = StorePrefix.LLMPrompt

    def __init__(
        self,
        key: Optional[str] = None,
        project: Optional[str] = None,
        model_artifact: Union[
            model_art.ModelArtifact, str
        ] = None,  # TODO support partial model uri
        prompt_template: Optional[list[dict]] = None,
        prompt_path: Optional[str] = None,
        prompt_legend: Optional[dict] = None,
        invocation_config: Optional[dict] = None,
        description: Optional[str] = None,
        target_path=None,
        **kwargs,
    ):
        llm_prompt_spec = LLMPromptArtifactSpec(
            prompt_template=prompt_template,
            prompt_path=prompt_path,
            prompt_legend=prompt_legend,
            model_artifact=model_artifact,
            invocation_config=invocation_config,
            target_path=target_path,
            description=description,
        )

        llm_metadata = ArtifactMetadata(
            key=key,
            project=project or "",
        )

        super().__init__(spec=llm_prompt_spec, metadata=llm_metadata, **kwargs)

    @property
    def spec(self) -> LLMPromptArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec: LLMPromptArtifactSpec):
        self._spec = self._verify_dict(spec, "spec", LLMPromptArtifactSpec)

    @property
    def model_artifact(self) -> Optional[model_art.ModelArtifact]:
        """
        Get the model artifact linked to this prompt artifact.
        """
        if self.spec._model_artifact:
            return self.spec._model_artifact
        if self.spec.model_uri:
            self.spec._model_artifact, _ = (
                mlrun.datastore.store_manager.get_store_artifact(self.spec.model_uri)
            )
            return self.spec._model_artifact
        return None

    def read_prompt(self) -> Optional[Union[str, list[dict]]]:
        """
        Read the prompt json from the artifact or if provided prompt template.
        @:param as_str: True to return the prompt string or a list of dicts.
        @:return prompt string or list of dicts
        """
        if self.spec.prompt_template:
            return self.spec.prompt_template
        if self.spec.target_path:
            with mlrun.datastore.store_manager.object(url=self.spec.target_path).open(
                mode="r"
            ) as p_file:
                try:
                    return json.load(p_file)
                except json.JSONDecodeError:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"Failed on decoding str in path "
                        f"{self.spec.target_path} expected file to contain a "
                        f"json format."
                    )

    def before_log(self):
        """
        Prepare the artifact before logging.
        This method is called before the artifact is logged.
        """
        if (
            self.spec.prompt_template
            and len(str(self.spec.prompt_template)) > MAX_PROMPT_LENGTH
        ):
            logger.debug(
                "Prompt string exceeds maximum length, saving to a temporary file."
            )
            with tempfile.NamedTemporaryFile(
                delete=False, mode="w", suffix=".json"
            ) as temp_file:
                temp_file.write(json.dumps(self.spec.prompt_template))
            self.spec.src_path = temp_file.name
            self.spec.prompt_template = None
            self._src_is_temp = True
        super().before_log()


class PlaceholderDefaultDict(defaultdict):
    def __missing__(self, key):
        return f"{{{key}}}"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import numpy as np
import plotly.graph_objects as go

import mlrun
from mlrun.artifacts import Artifact, PlotlyArtifact

from ..plan import MLPlanStages, MLPlotPlan
from ..utils import MLTypes, MLUtils


class FeatureImportancePlan(MLPlotPlan):
    """
    Plan for producing a feature importance.
    """

    _ARTIFACT_NAME = "feature-importance"

    def __init__(self):
        """
        Initialize a feature importance plan.

        An example of use can be seen at the Scikit-Learn docs here:
        https://scikit-learn.org/stable/auto_examples/ensemble/plot_forest_importances.html
        """
        super().__init__()

    def is_ready(self, stage: MLPlanStages, is_probabilities: bool) -> bool:
        """
        Check whether the plan is fit for production by the given stage and prediction probabilities. The feature
        importance is ready post training.

        :param stage:            The stage to check if the plan is ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not.

        :return: True if the plan is producible and False otherwise.
        """
        return stage == MLPlanStages.POST_FIT

    def produce(
        self, model: MLTypes.ModelType, x: MLTypes.DatasetType, **kwargs
    ) -> dict[str, Artifact]:
        """
        Produce the feature importance according to the given model and dataset ('x').

        :param model: Model to get its 'feature_importances_' or 'coef_' fields.
        :param x:     Input dataset the model trained on for the column labels.

        :return: The produced feature importance artifact in an artifacts dictionary.
        """
        # Get the importance score:
        importance_score = self._get_importance_score(model=model)
        if importance_score is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "This model cannot be used for Feature Importance plotting."
            )

        # Create a table of features and their importance along the assigned class:
        features = MLUtils.to_dataframe(dataset=x).columns
        targets = model.classes_ if hasattr(model, "classes_") else [0]
        if len(targets) != len(importance_score):
            targets = [0]

        # Create the figure:
        fig = go.Figure(
            [
                go.Bar(
                    x=class_importance_score,
                    y=[str(f) for f in features],
                    orientation="h",
                    name=str(target),
                )
                for class_importance_score, target in zip(importance_score, targets)
            ]
        )

        # Creating the artifact:
        self._artifacts[self._ARTIFACT_NAME] = PlotlyArtifact(
            key=self._ARTIFACT_NAME,
            figure=fig,
        )

        return self._artifacts

    @staticmethod
    def _get_importance_score(model: MLTypes.ModelType) -> Union[np.ndarray, None]:
        """
        Get the features importance score of the model. If the model do not hold the scores, None will be returned.

        :param model: Model to get its 'feature_importances_' or 'coef_' fields.

        :return: The feature importance scores if available and None otherwise.
        """
        # Set the score variable:
        importance_score: np.ndarray = None

        # Look for the importance score data:
        if hasattr(model, "feature_importances_"):
            # Tree-based feature importance:
            importance_score = model.feature_importances_
        elif hasattr(model, "feature_importance"):
            # Booster feature importance:
            importance_score = model.feature_importance()
        elif hasattr(model, "coef_"):
            # Coefficient-based importance:
            importance_score = model.coef_

        # If found, return it in two dimensions (first one for the classes amount):
        if importance_score is not None:
            if len(importance_score.shape) == 1:
                return [importance_score]
            return importance_score

        # If not found, check for inner estimator/s:
        if hasattr(model, "estimator_"):
            return FeatureImportancePlan._get_importance_score(model=model.estimator_)
        elif hasattr(model, "estimators_"):
            return np.vstack(
                [
                    FeatureImportancePlan._get_importance_score(model=estimator)
                    for estimator in model.estimators_
                ]
            )

        # No feature importance or coefficients are available for the given model:
        return None



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import random
import sys
from copy import deepcopy
from typing import Optional

import pandas as pd

from ..model import HyperParamOptions, RunObject, RunSpec
from ..utils import get_in

hyper_types = ["list", "grid", "random"]
default_max_iterations = 10
default_max_errors = 3


def get_generator(spec: RunSpec, execution, param_file_secrets: Optional[dict] = None):
    options = spec.hyper_param_options
    strategy = spec.strategy or options.strategy
    if not spec.is_hyper_job() or strategy == "custom":
        return None
    options.validate()
    hyperparams = spec.hyperparams
    param_file = spec.param_file or options.param_file
    if strategy and strategy not in hyper_types:
        raise ValueError(f"unsupported hyper params strategy  ({strategy})")

    if param_file and hyperparams:
        raise ValueError("hyperparams and param_file cannot be used together")

    options.selector = options.selector or spec.selector
    if options.selector:
        parse_selector(options.selector)

    obj = None
    if param_file:
        obj = execution.get_dataitem(param_file, secrets=param_file_secrets)

        # for csv files, it will contain a list of iterations to run.
        # its strategy must be a list, since grid and random strategies would yield a different
        # results than expected.
        # e.g.:
        # param file:
        # p1 | p2
        # 1  | 2
        # 3  | 4
        # ------
        # while strategy is grid, the number iteration would be the matrix of the two params instead of two.
        # thus, ignored.
        if obj.suffix == ".csv":
            if not strategy:
                strategy = "list"

            if strategy in ["grid", "random"]:
                raise ValueError(
                    "CSV param file cannot be used with grid or random strategy, "
                    "use a JSON file for parameters or leave empty."
                )
        elif not strategy or strategy in ["grid", "random"]:
            hyperparams = json.loads(obj.get())

    if not strategy or strategy == "grid":
        return GridGenerator(hyperparams, options)

    if strategy == "random":
        return RandomGenerator(hyperparams, options)

    if obj:
        df = obj.as_df()
    else:
        df = pd.DataFrame(hyperparams)
    return ListGenerator(df, options)


class TaskGenerator:
    def __init__(self, options: HyperParamOptions):
        self.options = options

    def use_parallel(self):
        return self.options.parallel_runs or self.options.dask_cluster_uri

    @property
    def max_errors(self):
        return self.options.max_errors or default_max_errors

    @property
    def max_iterations(self):
        return self.options.max_iterations or default_max_iterations

    def generate(self, run: RunObject):
        pass

    def eval_stop_condition(self, results) -> bool:
        if not self.options.stop_condition:
            return False
        return eval(self.options.stop_condition, {}, results)


class GridGenerator(TaskGenerator):
    def __init__(self, hyperparams, options=None):
        super().__init__(options)
        self.hyperparams = hyperparams

    def generate(self, run: RunObject):
        i = 0
        params = self.grid_to_list()
        max = len(next(iter(params.values())))

        while i < max:
            newrun = get_run_copy(run)
            param_dict = newrun.spec.parameters or {}
            for key, values in params.items():
                param_dict[key] = values[i]
            newrun.spec.parameters = param_dict
            newrun.metadata.iteration = i + 1
            i += 1
            yield newrun

    def grid_to_list(self):
        arr = {}
        lastlen = 1
        for pk, pv in self.hyperparams.items():
            for p in arr.keys():
                arr[p] = arr[p] * len(pv)
            expanded = []
            for i in range(len(pv)):
                expanded += [pv[i]] * lastlen
            arr[pk] = expanded
            lastlen = lastlen * len(pv)

        return arr


class RandomGenerator(TaskGenerator):
    def __init__(self, hyperparams: dict, options=None):
        super().__init__(options)
        self.hyperparams = hyperparams

    def generate(self, run: RunObject):
        i = 0

        while i < self.max_iterations:
            newrun = get_run_copy(run)
            param_dict = newrun.spec.parameters or {}
            params = {k: random.sample(v, 1)[0] for k, v in self.hyperparams.items()}
            for key, values in params.items():
                param_dict[key] = values
            newrun.spec.parameters = param_dict
            newrun.metadata.iteration = i + 1
            i += 1
            yield newrun


class ListGenerator(TaskGenerator):
    def __init__(self, df, options=None):
        super().__init__(options)
        self.df = df

    def generate(self, run: RunObject):
        i = 0
        for _, row in self.df.iterrows():
            newrun = get_run_copy(run)
            param_dict = newrun.spec.parameters or {}
            for key, values in row.to_dict().items():
                param_dict[key] = values
            newrun.spec.parameters = param_dict
            newrun.metadata.iteration = i + 1
            i += 1
            yield newrun


def get_run_copy(run: RunObject):
    newrun = deepcopy(run)
    newrun.spec.hyperparams = None
    newrun.spec.param_file = None
    newrun.spec.hyper_param_options = None
    return newrun


def parse_selector(criteria):
    idx = criteria.find(".")
    field = criteria
    if idx < 0:
        op = "max"
    else:
        op = criteria[:idx]
        field = criteria[idx + 1 :]
    if op not in ["min", "max"]:
        raise ValueError(
            f"illegal iteration selector {criteria}, "
            "selector format [min|max.]<result-name> e.g. max.accuracy"
        )
    return op, field


def selector(results: list, criteria):
    if not criteria:
        return 0, 0

    op, criteria = parse_selector(criteria)
    best_id = 0
    best_item = 0
    if op == "max":
        best_val = sys.float_info.min
    elif op == "min":
        best_val = sys.float_info.max

    i = 0
    for task in results:
        state = get_in(task, ["status", "state"])
        id = get_in(task, ["metadata", "iteration"])
        val = get_in(task, ["status", "results", criteria])
        if isinstance(val, str):
            try:
                val = float(val)
            except Exception:
                val = None
        if state != "error" and val is not None:
            if (op == "max" and val > best_val) or (op == "min" and val < best_val):
                best_id, best_item, best_val = id, i, val
        i += 1

    return best_item, best_id



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re
from typing import Optional, Union

from .data_types import ValueType, python_type_to_value_type
from .errors import MLRunRuntimeError, err_to_str
from .model import ModelObj


def _limited_string(value: str, max_size: int = 40):
    """
    Provide limited string size, typically for reporting original value
    in case of error (and for better identification of error location
    based on presenting part of original value)
    """
    return (
        value
        if (value is None) or (len(value) <= max_size)
        else value[:max_size] + "..."
    )


class Entity(ModelObj):
    """data entity (index)"""

    kind = "entity"

    def __init__(
        self,
        name: Optional[str] = None,
        value_type: Union[ValueType, str] = None,
        description: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
    ):
        """data entity (index key)

        :param name:        entity name
        :param value_type:  type of the entity, e.g. ValueType.STRING, ValueType.INT (default ValueType.STRING)
        :param description: test description of the entity
        :param labels:      a set of key/value labels (tags)
        """
        self.name = name
        self.description = description
        self.value_type = ValueType(value_type) if value_type else None
        if name and not value_type:
            self.value_type = ValueType.STRING
        self.labels = labels or {}

    def __eq__(self, other):
        return self.name == other.name


class Feature(ModelObj):
    _dict_fields = [
        "name",
        "description",
        "value_type",
        "dims",
        "default",
        "labels",
        "aggregate",
        "validator",
        "origin",
    ]

    def __init__(
        self,
        value_type: Union[ValueType, str] = None,
        dims: Optional[list[int]] = None,
        description: Optional[str] = None,
        aggregate: Optional[bool] = None,
        name: Optional[str] = None,
        validator=None,
        default: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
    ):
        """data feature

        Features can be specified manually or inferred automatically (during ingest/preview)

        :param value_type:  type of the feature. Use the ValueType constants library e.g. ValueType.STRING,
                            ValueType.INT (default ValueType.STRING)
        :param dims:        list of dimensions for vectors/tensors, e.g. [2, 2]
        :param description: text description of the feature
        :param aggregate:   is it an aggregated value
        :param name:        name of the feature
        :param validator:   feature validation policy
        :param default:     default value
        :param labels:      a set of key/value labels (tags). Labels can be used to filter featues, for example,
                            in the UI Feature store page.
        """
        self.name = name or ""
        if isinstance(value_type, ValueType):
            self.value_type = value_type
        elif value_type is not None:
            self.value_type = python_type_to_value_type(value_type)
        else:
            self.value_type = ValueType.STRING
        self.dims = dims
        self.description = description
        self.default = default
        self.labels = labels or {}
        self.aggregate = aggregate
        self.origin = None  # used to link the feature to the feature set origin (inside vector.status)
        self._validator = validator

    @property
    def validator(self):
        return self._validator

    @validator.setter
    def validator(self, validator):
        if isinstance(validator, dict):
            kind = validator.get("kind")
            validator = validator_kinds[kind].from_dict(validator)
        self._validator = validator


class BasicTypeValidator:
    def __init__(self):
        pass

    def check(self, value_type, value):
        return True, {}


class ConvertTypeValidator(BasicTypeValidator):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def check(self, value_type, value):
        ok, args = super().check(value_type, value)
        if ok:
            try:
                self.func(value)
            except Exception as err:
                return (
                    False,
                    {"message": err_to_str(err), "type": value_type},
                )
        return ok, args


class RangeTypeValidator(BasicTypeValidator):
    def __init__(self, min, max):
        super().__init__()
        self.min = min
        self.max = max

    def check(self, value_type, value):
        ok, args = super().check(value_type, value)
        if ok:
            try:
                if value < self.min:
                    return (
                        False,
                        {
                            "message": "Value is smaller than min range",
                            "type": value_type,
                            "min range": self.min,
                            "value": _limited_string(value),
                        },
                    )
                if value > self.max:
                    return (
                        False,
                        {
                            "message": "Value is greater than max range",
                            "type": value_type,
                            "max range": self.max,
                            "value": _limited_string(value),
                        },
                    )
            except Exception as err:
                return (
                    False,
                    {"message": err_to_str(err), "type": value_type},
                )

        return ok, args


# TODO: add addition validation for commented types
type_validator = {
    #   ValueType.BOOL: it does not make sense to do validation for BOOL (everything is True or False by default)
    ValueType.INT8: RangeTypeValidator(-128, 127),
    ValueType.INT16: RangeTypeValidator(-32768, 32767),
    ValueType.INT32: RangeTypeValidator(-2147483648, 2147483647),
    ValueType.INT64: RangeTypeValidator(-9223372036854775808, 9223372036854775807),
    ValueType.INT128: RangeTypeValidator(-math.pow(2, 127), math.pow(2, 127) - 1),
    ValueType.UINT8: RangeTypeValidator(0, 255),
    ValueType.UINT16: RangeTypeValidator(0, 65535),
    ValueType.UINT32: RangeTypeValidator(0, 4294967295),
    ValueType.UINT64: RangeTypeValidator(0, 18446744073709551615),
    ValueType.UINT128: RangeTypeValidator(0, math.pow(2, 128)),
    #   ValueType.FLOAT16: None,
    ValueType.FLOAT: ConvertTypeValidator(float),
    ValueType.DOUBLE: ConvertTypeValidator(float),
    #   ValueType.BFLOAT16: None,
    ValueType.BYTES: ConvertTypeValidator(bytes),
    #   ValueType.STRING: it does not make sense to do validation for STRING (everything is valid also '\x00', '\xff')
    #   ValueType.DATETIME: None,
    #   ValueType.BYTES_LIST: None,
    #   ValueType.STRING_LIST: None,
    #   ValueType.INT32_LIST: None,
    #   ValueType.INT64_LIST: None,
    #   ValueType.DOUBLE_LIST: None,
    #   ValueType.FLOAT_LIST: None,
    #   ValueType.BOOL_LIST: None,
}


class Validator(ModelObj):
    """Base validator"""

    kind = ""
    _dict_fields = ["kind", "check_type", "severity"]

    def __init__(
        self, check_type: Optional[bool] = None, severity: Optional[str] = None
    ):
        """Base validator

        example::

            from mlrun.features import Validator

            # Add validator to the feature 'bid' with check type
            quotes_set["bid"].validator = Validator(check_type=True, severity="info")

        :param check_type:  check feature type e.g. True, False
        :param severity:    severity name e.g. info, warning, etc.
        """
        self._feature = None
        self.check_type = check_type
        self.severity = severity

    def set_feature(self, feature: Feature):
        self._feature = feature

    def check(self, value):
        if self.check_type:
            if self._feature.value_type is not None:
                if self._feature.value_type in type_validator:
                    return type_validator[self._feature.value_type].check(
                        self._feature.value_type, value
                    )
        return True, {}


class MinMaxValidator(Validator):
    """Validate min/max value ranges"""

    kind = "minmax"
    _dict_fields = Validator._dict_fields + ["min", "max"]

    def __init__(
        self,
        check_type: Optional[bool] = None,
        severity: Optional[str] = None,
        min=None,
        max=None,
    ):
        """Validate min/max value ranges

        example::

            from mlrun.features import MinMaxValidator

            # Add validator to the feature 'bid', where valid
            # minimal value is 52
            quotes_set["bid"].validator = MinMaxValidator(min=52, severity="info")

        :param check_type:  check feature type e.g. True, False
        :param severity:    severity name e.g. info, warning, etc.
        :param min:         minimal valid size
        :param max:         maximal valid size
        """
        super().__init__(check_type, severity)
        self.min = min
        self.max = max

    def check(self, value):
        ok, args = super().check(value)
        if ok:
            try:
                if self.min is not None:
                    if value < self.min:
                        return (
                            False,
                            {
                                "message": "value is smaller than min",
                                "min": self.min,
                                "value": _limited_string(str(value)),
                            },
                        )
                if self.max is not None:
                    if value > self.max:
                        return (
                            False,
                            {
                                "message": "value is greater than max",
                                "max": self.max,
                                "value": _limited_string(str(value)),
                            },
                        )
            except Exception as err:
                return (
                    False,
                    {"message": err_to_str(err), "type": self.kind},
                )
        return ok, args


class MinMaxLenValidator(Validator):
    """Validate min/max length value ranges"""

    kind = "minmaxlen"
    _dict_fields = Validator._dict_fields + ["min", "max"]

    def __init__(
        self,
        check_type: Optional[bool] = None,
        severity: Optional[str] = None,
        min=None,
        max=None,
    ):
        """Validate min/max length value ranges

        example::

            from mlrun.features import MinMaxLenValidator

            # Add length validator to the feature 'ticker', where valid
            # minimal length is 1 and maximal length is 10
            quotes_set["ticker"].validator = MinMaxLenValidator(
                min=1, max=10, severity="info"
            )

        :param check_type:  check feature type e.g. True, False
        :param severity:    severity name e.g. info, warning, etc.
        :param min:         minimal valid length size
        :param max:         maximal valid length size
        """
        super().__init__(check_type, severity)
        self.min = min
        self.max = max

    def check(self, value):
        ok, args = super().check(value)
        if ok:
            try:
                if self.min is not None:
                    if len(value) < self.min:
                        return (
                            False,
                            {
                                "message": "Length value is smaller than min",
                                "min": self.min,
                                "length value": len(value),
                            },
                        )
                if self.max is not None:
                    if len(value) > self.max:
                        return (
                            False,
                            {
                                "message": "Length value is greater than max",
                                "max": self.max,
                                "length value": len(value),
                            },
                        )
            except Exception as err:
                return (
                    False,
                    {"message": err_to_str(err), "type": self.kind},
                )

        return ok, args


class RegexValidator(Validator):
    """Validate value based on regular expression"""

    kind = "regex"
    _dict_fields = Validator._dict_fields + ["regex"]

    def __init__(
        self,
        check_type: Optional[bool] = None,
        severity: Optional[str] = None,
        regex=None,
    ):
        """Validate value based on regular expression

        example::

            from mlrun.features import RegexValidator

            # Add regular expression validator to the feature 'name' and
            # expression '(\b[A-Za-z]{1}[0-9]{7}\b)' where valid values are
            # e.g. A1234567, z9874563, etc.
            quotes_set["name"].validator = RegexValidator(
                regex=r"(\b[A-Za-z]{1}[0-9]{7}\b)", severity="info"
            )

        :param check_type:  check feature type e.g. True, False
        :param severity:    severity name e.g. info, warning, etc.
        :param regex:       regular expression for validation
        """
        super().__init__(check_type, severity)
        self.regex = regex
        self.regex_compile = re.compile(self.regex) if self.regex else None

    def check(self, value):
        ok, args = super().check(value)
        if ok:
            try:
                if self.regex is not None:
                    if not re.fullmatch(self.regex_compile, value):
                        return (
                            False,
                            {
                                "message": "Value is not valid with regular expression",
                                "regexp": self.regex,
                                "value": _limited_string(str(value)),
                            },
                        )
            except Exception as err:
                return (
                    False,
                    {"message": err_to_str(err), "type": self.kind},
                )
        return ok, args

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: Optional[dict] = None
    ):
        new_obj = super().from_dict(
            struct=struct, fields=fields, deprecated_fields=deprecated_fields
        )
        if hasattr(new_obj, "regex"):
            new_obj.regex_compile = re.compile(new_obj.regex) if new_obj.regex else None
        else:
            raise MLRunRuntimeError(
                f"Object with type {type(new_obj)} "
                f"have to contain `regex` attribute"
            )
        return new_obj


validator_kinds = {
    "": Validator,
    "minmax": MinMaxValidator,
    "minmaxlen": MinMaxLenValidator,
    "regex": RegexValidator,
}



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import tensorflow as tf

import mlrun
import mlrun.common.constants as mlrun_constants

from .callbacks import MLRunLoggingCallback, TensorboardLoggingCallback
from .mlrun_interface import TFKerasMLRunInterface
from .model_handler import TFKerasModelHandler
from .model_server import TFKerasModelServer
from .utils import TFKerasTypes, TFKerasUtils


def apply_mlrun(
    model: tf.keras.Model = None,
    model_name: Optional[str] = None,
    tag: str = "",
    model_path: Optional[str] = None,
    model_format: Optional[str] = None,
    save_traces: bool = False,
    modules_map: Optional[Union[dict[str, Union[None, str, list[str]]], str]] = None,
    custom_objects_map: Optional[Union[dict[str, Union[str, list[str]]], str]] = None,
    custom_objects_directory: Optional[str] = None,
    context: mlrun.MLClientCtx = None,
    auto_log: bool = True,
    tensorboard_directory: Optional[str] = None,
    mlrun_callback_kwargs: Optional[dict[str, Any]] = None,
    tensorboard_callback_kwargs: Optional[dict[str, Any]] = None,
    use_horovod: Optional[bool] = None,
    **kwargs,
) -> TFKerasModelHandler:
    """
    Wrap the given model with MLRun's interface providing it with mlrun's additional features.

    :param model:                       The model to wrap. Can be loaded from the model path given as well.
    :param model_name:                  The model name to use for storing the model artifact. If not given, the
                                        tf.keras.Model.name will be used.
    :param tag:                         The model's tag to log with.
    :param model_path:                  The model's store object path. Mandatory for evaluation (to know which model to
                                        update). If model is not provided, it will be loaded from this path.
    :param model_format:                The format to use for saving and loading the model. Should be passed as a
                                        member of the class 'ModelFormats'.
    :param save_traces:                 Whether or not to use functions saving (only available for the 'SavedModel'
                                        format) for loading the model later without the custom objects dictionary. Only
                                        from tensorflow version >= 2.4.0. Using this setting will increase the model
                                        saving size.
    :param modules_map:                 A dictionary of all the modules required for loading the model. Each key is a
                                        path to a module and its value is the object name to import from it. All the
                                        modules will be imported globally. If multiple objects needed to be imported
                                        from the same module a list can be given. The map can be passed as a path to a
                                        json file as well. For example:

                                        .. code-block:: python

                                            {
                                                "module1": None,  # import module1
                                                "module2": ["func1", "func2"],  # from module2 import func1, func2
                                                "module3.sub_module": "func3",  # from module3.sub_module import func3
                                            }

                                        If the model path given is of a store object, the modules map will be read from
                                        the logged modules map artifact of the model.
    :param custom_objects_map:          A dictionary of all the custom objects required for loading the model. Each key
                                        is a path to a python file and its value is the custom object name to import
                                        from it. If multiple objects needed to be imported from the same py file a list
                                        can be given. The map can be passed as a path to a json file as well. For
                                        example:

                                        .. code-block:: python

                                            {
                                                "/.../custom_optimizer.py": "optimizer",
                                                "/.../custom_layers.py": ["layer1", "layer2"],
                                            }

                                        All the paths will be accessed from the given 'custom_objects_directory',
                                        meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                        If the model path given is of a store object, the custom objects map will be
                                        read from the logged custom object map artifact of the model.
                                        Notice: The custom objects will be imported in the order they came in this
                                        dictionary (or json). If a custom object is depended on another, make sure to
                                        put it below the one it relies on.
    :param custom_objects_directory:    Path to the directory with all the python files required for the custom objects.
                                        Can be passed as a zip file as well (will be extracted during the run before
                                        loading the model). If the model path given is of a store object, the custom
                                        objects files will be read from the logged custom object artifact of the model.
    :param context:                     MLRun context to work with. If no context is given it will be retrieved via
                                        'mlrun.get_or_create_ctx(None)'
    :param auto_log:                    Whether or not to apply MLRun's auto logging on the model. Default: True.
    :param tensorboard_directory:       If context is not given, or if wished to set the directory even with context,
                                        this will be the output for the event logs of tensorboard. If not given, the
                                        'tensorboard_dir' parameter will be tried to be taken from the provided context.
                                        If not found in the context, the default tensorboard output directory will be:
                                        /User/.tensorboard/<PROJECT_NAME> or if working on local, the set artifacts
                                        path.
    :param mlrun_callback_kwargs:       Key word arguments for the MLRun callback. For further information see the
                                        documentation of the class 'MLRunLoggingCallback'. Note that both 'context'
                                        and 'auto_log' parameters are already given here.
    :param tensorboard_callback_kwargs: Key word arguments for the tensorboard callback. For further information see
                                        the documentation of the class 'TensorboardLoggingCallback'. Note that both
                                        'context' and 'auto_log' parameters are already given here.
    :param use_horovod:                 Whether or not to use horovod - a distributed training framework. Default:
                                        None, meaning it will be read from context if available and if not - False.

    :return: The model with MLRun's interface.
    """
    # Get parameters defaults:
    # # Context:
    if context is None:
        context = mlrun.get_or_create_ctx(TFKerasMLRunInterface.DEFAULT_CONTEXT_NAME)
    # # Use horovod:
    if use_horovod is None:
        use_horovod = (
            context.labels.get(mlrun_constants.MLRunInternalLabels.kind, "") == "mpijob"
            if context is not None
            else False
        )

    # Create a model handler:
    model_handler_kwargs = (
        kwargs.pop("model_handler_kwargs") if "model_handler_kwargs" in kwargs else {}
    )
    handler = TFKerasModelHandler(
        model_name=model_name,
        model_path=model_path,
        model=model,
        model_format=model_format,
        save_traces=save_traces,
        context=context,
        modules_map=modules_map,
        custom_objects_map=custom_objects_map,
        custom_objects_directory=custom_objects_directory,
        **model_handler_kwargs,
    )

    # Load the model if it was not provided:
    if model is None:
        handler.load()
        model = handler.model

    # Add MLRun's interface to the model:
    TFKerasMLRunInterface.add_interface(obj=model)

    # Initialize horovod if needed:
    if use_horovod is True:
        model.use_horovod()

    # Add auto-logging if needed:
    if auto_log:
        # Set the kwargs dictionaries defaults:
        mlrun_callback_kwargs = (
            {} if mlrun_callback_kwargs is None else mlrun_callback_kwargs
        )
        tensorboard_callback_kwargs = (
            {} if tensorboard_callback_kwargs is None else tensorboard_callback_kwargs
        )
        # Add the logging callbacks with the provided parameters:
        model.add_logging_callback(
            logging_callback=MLRunLoggingCallback(
                context=context,
                model_handler=handler,
                log_model_tag=tag,
                auto_log=auto_log,
                **mlrun_callback_kwargs,
            )
        )
        model.add_logging_callback(
            logging_callback=TensorboardLoggingCallback(
                context=context,
                tensorboard_directory=tensorboard_directory,
                auto_log=auto_log,
                **tensorboard_callback_kwargs,
            )
        )

    return handler
